/**
  *  \file client/cscr-ch.cc
  *  \brief Control Screens -- Starchart
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/cscreen.h"
#include "gfx/palette.h"
#include "gfx/clip.h"
#include "gfx/gfx.h"
#include "game/lock.h"

/****************************** WChartWidget *****************************/

/*! \class WChartWidget

    This just displays a static piece of starchart, without any user
    interaction. It can be made invalid, in this case it will not
    display anything. */

/** Constructor. Call setCenterPosition() to set position to display. */
WChartWidget::WChartWidget()
    : is_valid(false)
{
    viewport.center = GPoint(2000, 2000);
    viewport.mult = viewport.divi = 1;
    viewport.vec_id = 0;
}

WChartWidget::~WChartWidget()
{ }

void
WChartWidget::draw(GfxCanvas& can, UpdateInfo& ui)
{
    if (!isValid()) {
        drawSolidBar(can, getExtent(), standard_colors[COLOR_BLACK]);
        drawPost(can);
    } else {
        viewport.view = getExtent();
        viewport.init();            // so that drawPre/drawPost can use it

        can.drawBar(getExtent(), standard_colors[COLOR_BLACK], standard_colors[COLOR_BLACK],
                    SOLID_FILL, GFX_ALPHA_OPAQUE);
        drawPre(can);
        viewport.draw(can, getChartOpts(true, viewport.mult, viewport.divi),
                      getDisplayedTurn(),
                      getDisplayedTurn().getChartBBox());
        drawPost(can);
    }
    ui.update(getExtent());
}

/** Set position of the center of this chart widget. Will update the
    display. This will also mark the widget valid.

    Derived classes can override this function to do their own
    processing, but must call this inherited version for the change to
    take effect. */
void
WChartWidget::setCenterPosition(GPoint pt)
{
    if (is_valid && pt == viewport.center)
        return;
    is_valid = true;
    viewport.center = pt;
    drawWidget(false);
}

/** Set chart widget invalid. This will remove the map display and
    only display a black area. */
void
WChartWidget::setInvalid()
{
    if (!is_valid)
        return;
    is_valid = true;
    drawWidget(false);
}

/** Pre-draw hook. Override in descendant classes to draw something
    below the map. This is only called if the starchart viewport
    actually is valid. */
void
WChartWidget::drawPre(GfxCanvas& /*can*/)
{ }

/** Post-draw hook. Override in descendant classes to draw something
    atop the map. This is also called with invalid starchart viewport,
    so it can draw a notice why the map is not displayed this time. */
void
WChartWidget::drawPost(GfxCanvas& /*can*/)
{ }

/** Zoom in.
    \return true iff zoom factor actually changed. */
bool
WChartWidget::zoomIn()
{
    if (viewport.zoomIn(getDisplayedTurn().getChartBBox())) {
        drawWidget(false);
        return true;
    }
    return false;
}

/** Zoom out.
    \return true iff zoom factor actually changed. */
bool
WChartWidget::zoomOut()
{
    if (viewport.zoomOut(getDisplayedTurn().getChartBBox())) {
        drawWidget(false);
        return true;
    }
    return false;
}

/************************** WScannerChartWidget **************************/

/*! \class WScannerChartWidget
    \brief Chart Widget for Control Screen Scanners

    This contains a standard chart widget with a movable scanner beam. */

/** Constructor. */
WScannerChartWidget::WScannerChartWidget()
    : scan(getViewport().center)
{ }

WScannerChartWidget::~WScannerChartWidget()
{ }

void
WScannerChartWidget::drawScannerXor(GfxCanvas& can)
{
    GfxXorFilter      xor_flt(can);
    GfxClipRectFilter can1(getExtent(), xor_flt);

    drawScanner(can1);
}

/** Draw end of scanner beam, large version.
    \param can Canvas
    \param parm Parameters
    \param pt Screen coordinates */
void
WScannerChartWidget::drawEndBig(GfxCanvas& can, GfxParameters& parm, GfxPoint pt)
{
    drawVLine(can, parm, pt.x-3, pt.y-2, pt.y+2);
    drawVLine(can, parm, pt.x+3, pt.y-2, pt.y+2);
    drawHLine(can, parm, pt.x-2, pt.y-3, pt.x+2);
    drawHLine(can, parm, pt.x-2, pt.y+3, pt.x+2);
}

/** Draw end of scanner beam, small version.
    \param can Canvas
    \param parm Parameters
    \param pt Screen coordinates  */
void
WScannerChartWidget::drawEndSmall(GfxCanvas& can, GfxParameters& parm, GfxPoint pt)
{
    can.drawPixel(pt.x+1, pt.y, parm.color, GFX_ALPHA_OPAQUE);
    can.drawPixel(pt.x-1, pt.y, parm.color, GFX_ALPHA_OPAQUE);
    can.drawPixel(pt.x, pt.y+1, parm.color, GFX_ALPHA_OPAQUE);
    can.drawPixel(pt.x, pt.y-1, parm.color, GFX_ALPHA_OPAQUE);
}

/** Draw scanner beam. */
void
WScannerChartWidget::drawScanner(GfxCanvas& can)
{
    GfxParameters parm;
    parm.color = standard_colors[COLOR_YELLOW];
    const GChartViewport& vp = getViewport();
    void (*f)(GfxCanvas&,GfxParameters&,GfxPoint) = vp.mult >= vp.divi ? drawEndBig : drawEndSmall;

    GfxPoint cen = vp.getCenter();
    GfxPoint end = vp.scaleXY(scan);

    f(can, parm, cen);
    if (end != cen) {
        drawLine(can, parm, cen.x, cen.y, end.x, end.y);
        f(can, parm, end);
    }
}

/** Draw below starchart. This draws the center cross. */
void
WScannerChartWidget::drawPre(GfxCanvas& can)
{
    const GfxPoint m = getViewport().getCenter();

    GfxParameters parm;
    parm.color = standard_colors[COLOR_BLUEBLACK];

    drawHLine(can, parm, m.x - 10, m.y,      m.x + 11);
    drawVLine(can, parm, m.x,      m.y - 10, m.y + 11);
}

/** Draw above starchart. This draws the scanner. */
void
WScannerChartWidget::drawPost(GfxCanvas& can)
{
    if (isValid())
        drawScannerXor(can);
}

/** Clip scanner position.
    \return pt modified such that it is in the visible range
    \todo maybe move to WChartWidget? */
GPoint
WScannerChartWidget::clipScannerPosition(GPoint pt) const
{
    const GChartViewport& vp = getViewport();
    const GPoint m           = vp.center;
    const int xw = vp.unscale(extent.w) / 2;
    const int yw = vp.unscale(extent.h) / 2;
    if (pt.x < m.x - xw)
        pt.x = m.x - xw;
    if (pt.x > m.x + xw)
        pt.x = m.x + xw;
    if (pt.y < m.y - yw)
        pt.y = m.y - yw;
    if (pt.y > m.y + yw)
        pt.y = m.y + yw;
    return pt;
}

/** Set scanner position.
    \param pt      point
    \param limit   true to limit actual value to visible range,
                   false to take new value without checking */
void
WScannerChartWidget::setScannerPosition(GPoint pt)
{
    if (pt != scan) {
        if (hasState(st_Exposed) && !hasState(st_WantRedraw)) {
            /* widget is visible and up-to-date */
            ObjectHolder h;
            GfxCanvas& can = owner->getCanvasFor(this, h);
            can.lock();
            drawScannerXor(can);
            scan = pt;
            drawScannerXor(can);
            can.unlock();
            can.update(extent);
            sig_moved.raise(getId());
        } else {
            /* widget is either not visible, or will be redrawn soon,
               so we don't need to do anything */
            scan = pt;
            sig_moved.raise(getId());
        }
    }
}

/** Change scanner center position. This will also reset the scanner
    position to point to the center. */
void
WScannerChartWidget::setCenterPosition(GPoint pt)
{
    WChartWidget::setCenterPosition(pt);
    setScannerPosition(pt);
}

void
WScannerChartWidget::userMoveScannerTo(GPoint pt)
{
    setScannerPosition(pt);
}

/** Move according to a keypress. */
void
WScannerChartWidget::moveKey(int dx, int dy)
{
    select();
    GPoint pt = getScannerPosition();
    userMoveScannerTo(clipScannerPosition(GPoint(pt.x + dx, pt.y + dy)));
}

/** Handle keypress.
    \param k key
    \return true iff key handled */
bool
WScannerChartWidget::handleKey(UIEvent::Key k)
{
    switch (k.code) {
     case SDLK_LEFT:
        moveKey(k.arg ? -k.arg : -10, 0);
        return true;
     case SDLK_RIGHT:
        moveKey(k.arg ? k.arg : 10, 0);
        return true;
     case SDLK_UP:
        moveKey(0, k.arg ? k.arg : 10);
        return true;
     case SDLK_DOWN:
        moveKey(0, k.arg ? -k.arg : -10);
        return true;
     case ss_Shift + SDLK_LEFT:
        moveKey(-1, 0);
        return true;
     case ss_Shift + SDLK_RIGHT:
        moveKey(1, 0);
        return true;
     case ss_Shift + SDLK_UP:
        moveKey(0, 1);
        return true;
     case ss_Shift + SDLK_DOWN:
        moveKey(0, -1);
        return true;
     case ss_Ctrl + SDLK_LEFT:
        moveKey(-100, 0);
        return true;
     case ss_Ctrl + SDLK_RIGHT:
        moveKey(100, 0);
        return true;
     case ss_Ctrl + SDLK_UP:
        moveKey(0, 100);
        return true;
     case ss_Ctrl + SDLK_DOWN:
        moveKey(0, -100);
        return true;
     case '+':
        if (hasState(sc_Zoomable))
            zoomIn();
        return true;
     case '-':
        if (hasState(sc_Zoomable))
            zoomOut();
        return true;
     case ' ':
     case SDLK_RETURN:
     case ss_Ctrl + ' ':
     case ss_Ctrl + SDLK_RETURN:
        {
            const GChartViewport& vp = getViewport();
            GLockData locker(getScannerPosition(),
                             GPoint(vp.unscaleX(extent.x),
                                    vp.unscaleY(extent.y + extent.h - 1)),
                             GPoint(vp.unscaleX(extent.x + extent.w - 1),
                                    vp.unscaleY(extent.y)),
                             (k.code & ss_Ctrl) != 0);
            findItem(getDisplayedTurn(), locker,
                     (k.code & ~ss_Ctrl) == ' ' ? lm_Left : lm_Right);
            userMoveScannerTo(locker.getPoint());
        }
        return true;

        /* z
           m c-m */
    }
    return false;
}

bool
WScannerChartWidget::handleEvent(const UIEvent& e, bool second_pass)
{
    /* nothing selected -> no input accepted */
    if (!isValid())
        return WChartWidget::handleEvent(e, second_pass);

    if (e.type == ev_Mouse && extent.contains(e.mouse.x, e.mouse.y)) {
        const GChartViewport& vp = getViewport();
        select();
        if (e.mouse.b || hasState(sc_KeyboardMode)) {
            if (e.mouse.b & mb_Double) {
                sig_dblclick.raise(getId());
            } else {
                GLockData locker(GPoint(vp.unscaleX(e.mouse.x),
                                        vp.unscaleY(e.mouse.y)),
                                 GPoint(vp.unscaleX(extent.x),
                                        vp.unscaleY(extent.y + extent.h - 1)),
                                 GPoint(vp.unscaleX(extent.x + extent.w - 1),
                                        vp.unscaleY(extent.y)),
                                 0);
                if (e.mouse.b & mb_Left)
                    findItem(getDisplayedTurn(), locker, lm_Left);
                else if (e.mouse.b & mb_Right)
                    findItem(getDisplayedTurn(), locker, lm_Right);

                userMoveScannerTo(locker.getPoint());
            }
        }
        return true;
    } else if (e.type == ev_Key && hasState(sc_KeyboardMode | sc_Keyboard)) {
        return handleKey(e.key);
    } else {
        return WChartWidget::handleEvent(e, second_pass);
    }
}

void
WScannerChartWidget::getLayoutInfo(LayoutInfo& info)
{
    // FIXME? Probably preferred size should be different
    info.flags = li_Flexible;
    info.min_size = info.pref_size = GfxPoint(100,100);
}
