/**
  *  \file client/dlg-structure.cc
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dlg-structure.h"
#include "game/action.h"
#include "util/nls.h"
#include "gfx/font.h"
#include "gfx/palette.h"
#include "cpluslib/holder.h"
#include "ui/commands.h"
#include "ui/layout.h"
#include "game/global.h"
#include "ui/msgbox.h"
#include "ui/skin.h"

WPlanetStructureWidget::WPlanetStructureWidget(GPlanetBuildStructuresAction& action,
                                               TPlanetaryStructure structure)
    : action(action), structure(structure)
{
    conn = action.sig_changed.add(static_cast<UIWidget*>(this),
                                  (void (UIWidget::*)()) &UIWidget::drawWidget);
}

WPlanetStructureWidget::~WPlanetStructureWidget()
{ }

void
WPlanetStructureWidget::add(int arg)
{
    action.addStructuresLimitCash(structure, arg ? arg : 1);
}

void
WPlanetStructureWidget::remove(int arg)
{
    action.addStructuresLimitCash(structure, arg ? -arg : -1);
}

void
WPlanetStructureWidget::draw(GfxCanvas& can, UpdateInfo& ui)
{
    static const char*const names[] = {
        N_("Mineral Mines:"),
        N_("Factories:"),
        N_("Defense Posts:"),
        N_("Starbase Defense:")
    };
    static const char*const costs[] = {
        // FIXME: I don't really like to hardcode these strings here
        // and the costs in GPlanetBuildStructuresAction. Better
        // solution anyone?
        N_("4 mc + 1 supply"),
        N_("3 mc + 1 supply"),
        N_("10 mc + 1 supply"),
        N_("10 mc + 1 Duranium")
    };
    const int x = getExtent().x, y = getExtent().y,
        y2 = y + font_heights[FONT_TITLE]+5, line = font_heights[FONT_NORMAL];

    UISkin& s = getSkin();
    s.startDraw();
    s.setSolid(false);
    s.setColor(UIColor::tc_White);
    s.outText(can, FONT_TITLE, x, y, _(names[structure]));

    s.setColor(UIColor::tc_Static);
    s.outText(can, FONT_NORMAL, x, y2,        _("Amount:"));
    s.outText(can, FONT_NORMAL, x, y2+line,   _("Auto-B. Goal:"));
    s.outText(can, FONT_NORMAL, x, y2+2*line, _("Maximum:"));
    s.outText(can, FONT_NORMAL, x, y2+3*line, _("Cost:"));

    s.setColor(UIColor::tc_Green);
    s.outText(can, FONT_NORMAL, x+45, y2+3*line, _(costs[structure]));

    s.setSolid();
    s.getParameters().just_x = 2;
    s.outTextF(can, FONT_NORMAL, x+110, y2,        40, itoa(action.getAmount(structure)));
    s.outTextF(can, FONT_NORMAL, x+110, y2+2*line, 40, itoa(action.getMax(structure)));

    ui.update(getExtent());
}

bool
WPlanetStructureWidget::handleEvent(const UIEvent& e, bool second_pass)
{
    return UIWidget::handleEvent(e, second_pass);
}

void
WPlanetStructureWidget::getLayoutInfo(LayoutInfo& info)
{
    info.flags = 0;
    info.pref_size = info.min_size = GfxPoint(115,
                                              3*font_heights[FONT_NORMAL]
                                              + font_heights[FONT_TITLE] + 5);
}

/************************* WPlanetStructureGroup *************************/

WPlanetStructureGroup::WPlanetStructureGroup(int id,
                                             GPlanetBuildStructuresAction& action,
                                             TPlanetaryStructure structure)
    : UIFocusGroup(id),
      psw(action, structure), add_btn(0, "+", '+', UIButton::aNothing),
      sub_btn(0, "-", '-', UIButton::aNothing),
      structure(structure)
{
    add(psw);
    add(add_btn);
    add(sub_btn);
    add_btn.sig_fired.add(&psw, &WPlanetStructureWidget::add);
    sub_btn.sig_fired.add(&psw, &WPlanetStructureWidget::remove);
}

WPlanetStructureGroup::~WPlanetStructureGroup()
{ }

void
WPlanetStructureGroup::getLayoutInfo(LayoutInfo& info)
{
    info.flags = 0;
    info.min_size = info.pref_size = GfxPoint(315, 92+10); // FIXME: magic numbers
}

void
WPlanetStructureGroup::setExtent(GfxRect newpos, bool by_layout)
{
    UIFocusGroup::setExtent(newpos, by_layout);

    psw.setExtent(GfxRect(newpos.x + 120, newpos.y + 5, 165, 92), by_layout);
    add_btn.setExtent(GfxRect(newpos.x + 285, newpos.y +  5, 25, 25), by_layout);
    sub_btn.setExtent(GfxRect(newpos.x + 285, newpos.y + 35, 25, 25), by_layout);
}

void
WPlanetStructureGroup::drawInterior(GfxCanvas& can, UpdateInfo& ui)
{
    static const char*const resid[] = {
        "planet.mine",
        "planet.factory",
        "planet.defense",
        "base.defense"
    };

    if (Ptr<GfxPixmap> pix = requestPixmap(ResId(resid[structure])))
        pix->blitSized(can, extent.x+5, extent.y+5, 105, 93);
    ui.update(GfxRect(extent.x+5, extent.y+5, 105, 93));
    UIFocusGroup::drawInterior(can, ui);
}

void
doPlanetBuildScreen(GPlanet& planet, int start_screen)
{
    // FIXME: implement actual version from PCC 1.x
    // FIXME: use GPlanetBuildStructuresAction::getError()
    GMultiTransaction mt;
    GPlanetTransfer& transfer =
        mt.addNewSubtransaction(new GPlanetTransfer(planet));
    GPlanetBuildStructuresAction& action =
        mt.addNewSubtransaction(new GPlanetBuildStructuresAction(planet, transfer));

    if (GError e = mt.getError()) {
        messageBox(e.getUserError(),
                   _("Build Structures"));
        return;
    }

    ObjectHolder h;             // ATTN: widgets must die before the action

    UIWindow win(_("Build Structures"), UIWindow::BLUE_WINDOW, COLOR_DARK);
    win.setLayout(UIVBoxLayout::instance5);

    int id = 100;
    if (start_screen == 0 || start_screen == 1)
        win.add(h.add(new WPlanetStructureGroup(id++, action, str_Mines)));
    if (start_screen == 0)
        win.add(h.add(new WPlanetStructureGroup(id++, action, str_Factories)));
    if (start_screen == 0 || start_screen == 2)
        win.add(h.add(new WPlanetStructureGroup(id++, action, str_Defense)));
    if (start_screen == 2 && planet.hasBase())
        win.add(h.add(new WPlanetStructureGroup(id++, action, str_BaseDefense)));

    win.add(h.add(new UIFocusIterator(100, id-1, UIFocusIterator::fi_Vert | UIFocusIterator::fi_Tab)));

    UIGroup g(UIHBoxLayout::instance5);
    g.add(h.add(new UISpacer()));
    g.add(h.add(new UIButton(cm_Enter, _("OK"), SDLK_RETURN, UIButton::aStop)));
    g.add(h.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE, UIButton::aStop)));
    win.add(g);
    win.pack();
    ui_root->centerWidget(win);
    ui_root->add(win);
    int rv = ui_root->run();
    ui_root->close(win);
    if (rv == cm_Enter && !mt.getError()) {
        mt.commit();
    } else {
        mt.cancel();
    }
    getDisplayedTurn().doUpdates();
}
