/**
  *  \file main.cc
  *  \brief Client Main Module
  *
  *  This file contains the main function for the planets client.
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#include <cstdlib>
#include "arch/platform.h"
#include "client/chart.h"
#include "client/chartusr.h"
#include "client/playerscr.h"
#include "cpluslib/format.h"
#include "game/config.h"
#include "game/fcode.h"
#include "game/global.h"
#include "game/load.h"
#include "game/mission.h"
#include "game/player.h"
#include "game/reg.h"
#include "game/save.h"
#include "game/spec.h"
#include "gfx/gfx.h"
#include "gfx/init.h"
#include "gfx/palette.h"
#include "ui/msgbox.h"
#include "ui/uiglobal.h"
#include "util/cmdline.h"
#include "util/console.h"
#include "util/debug.h"
#include "util/nls.h"
#include "util/recode.h"
#include "version.h"
#include "client/gdchooser.h"
#include "ui/commands.h"
#include "ui/widgets.h"
#include "util/confpars.h"

namespace {
    /** Option Parser for pcc-v2. */
    class PCC2OptionParser : public StandardCommandLineParser {
        enum { MAX_NON_OPTIONS = 2 };
        int num_nonoptions;
        int player;
        string_t nonopts[MAX_NON_OPTIONS];
        static void help(StandardCommandLineParser&, const Option&);
        static void setCharset(StandardCommandLineParser&, const Option&);
     public:
        PCC2OptionParser();
        void handleNonoption(const string_t& str);
        void handleError(const string_t& msg);
        int getNumNonoptions() const { return num_nonoptions; }
        void parseFinished();

        string_t getGameDirName() const { return nonopts[0]; }
        string_t getRootDirName() const { return nonopts[1]; }
        int getPlayerId() const { return player; }

        void setGameDirName(string_t gd)
            { nonopts[0] = gd; }
        void setPlayerId(int pid)
            { player = pid; }

        bool chooseGame(Ptr<Directory> dir);
    };

    // FIXME: PCC2GameChooser and PCC2GameChooserWindow are almost
    // complete copies of WVcrGameChooser and WVcrGameChooserWindow.
    class PCC2GameChooser : public WGameRaceChooser {
        WGamedirChooser& gdc;
     protected:
        void handleMovement();
     public:
        PCC2GameChooser(int id, WGamedirChooser& gdc);
        void scan(Ptr<Directory> dir);
    };

    class PCC2GameChooserWindow : public UIWindow {
        WGamedirChooser    dir_chooser;
        PCC2GameChooser    game_chooser;
        ObjectHolder       holder;
        PCC2OptionParser&  parser;
     public:
        PCC2GameChooserWindow(Ptr<Directory> dir, bool oneup, PCC2OptionParser& parser);
        ~PCC2GameChooserWindow();
        bool handleEvent(const UIEvent& e, bool second_pass);
        void maybeStop(int);
        void switchTo(int id);
    };
}

/** Initialize. */
PCC2OptionParser::PCC2OptionParser()
    : num_nonoptions(0), player(0)
{
    static const Option options[] = {
        { "help",    help,       0, false,
          N_(" -help              Show help screen") },
        { "charset", setCharset, 0, true,
          N_(" -charset EXT[:INT] Set character set translation") }
    };
    addOptionSet(options, countof(options));
}

/** Handle Non-Options. The first non-option is the game directory
    name, the second (optional) is the root directory. Player number
    can stand anywhere. */
void
PCC2OptionParser::handleNonoption(const string_t& str)
{
    if (!player) {
        int x = std::atoi(str.c_str());
        if (x > 0 && x <= NUM_PLAYERS) {
            player = x;
            return;
        }
    }
    if (num_nonoptions >= MAX_NON_OPTIONS)
        handleError(_("Too many parameters specified."));
    nonopts[num_nonoptions++] = str;
}

/** Show help screen. */
void
PCC2OptionParser::help(StandardCommandLineParser& par,
                       const StandardCommandLineParser::Option&)
{
    console << _("Usage: pcc-v2 [-options] gamedir [playerid] [rootdir]\n"
                 "\n"
                 "Valid options:\n");
    par.printHelp(console);
    console << _("\n(c) 2001-2004 Stefan Reuther <Streu@gmx.de>\n");
    Platform::showOutput();
    std::exit(0);
}

void
PCC2OptionParser::setCharset(StandardCommandLineParser& par,
                             const StandardCommandLineParser::Option&)
{
    string_t cs = par.getArg();
    string_t intrn = "latin1";
    strSplit(cs, cs, intrn, ":");
    CharacterSet ics = Recoder::getCharsetByName(intrn);
    CharacterSet ecs = Recoder::getCharsetByName(cs);
    if (!ics.isValid())
        console.write(LOG_WARN, format(_("Character set `%s' is unknown"), intrn));
    if (!ecs.isValid())
        console.write(LOG_WARN, format(_("Character set `%s' is unknown"), cs));
    Recoder::setCharsets(ics, ecs);

}

/** Error handler. Instead of throwing an exception, show a message
    and bail out. */
void
PCC2OptionParser::handleError(const string_t& str)
{
    console.write(LOG_ERROR, str);
    console.write(LOG_ERROR, _("Try `pcc-v2 --help' for help."));
    Platform::showOutput();
    std::exit(1);
}

/** Parsing finished. Postprocess input. */
void
PCC2OptionParser::parseFinished()
{
    if (getNumNonoptions() < 2)
        nonopts[1] = Platform::getDataDirectory();
}

bool
PCC2OptionParser::chooseGame(Ptr<Directory> dir)
{
    PCC2GameChooserWindow win(dir, false, *this);
    ui_root->centerWidget(win);
    ui_root->add(win);
    int rv = ui_root->run();
    ui_root->close(win);
    return rv == cm_Enter;
}

/** Initialize.
    \param id      widget id
    \param gdc     associated game directory chooser */
PCC2GameChooser::PCC2GameChooser(int id, WGamedirChooser& gdc)
    : WGameRaceChooser(id), gdc(gdc)
{
    gdc.sig_moved.add(this, &PCC2GameChooser::handleMovement);
}

void
PCC2GameChooser::scan(Ptr<Directory> dir)
{
    GDirectoryOverview ov;
    ov.scan(*dir);

    for (int i = 1; i <= NUM_PLAYERS; ++i) {
        if (ov.getPlayerFlags(i) & GDirectoryOverview::have_RST) {
            if (ov.getPlayerFlags(i) & GDirectoryOverview::have_TRN)
                setInfo(i, true, "RST + TRN");
            else
                setInfo(i, true, "RST");
        } else if (ov.getPlayerFlags(i) & GDirectoryOverview::have_Gamedir) {
            setInfo(i, true, "unpacked");
        }
    }
}

/** Update after directory changed. */
void
PCC2GameChooser::handleMovement()
{
    setDirectory(gdc.getSubdirectory());
}

/** Initialize.
    \param dir     initial directory
    \param oneup   false: display directory as is;
                   true: display dir's parent, place cursor on dir */
PCC2GameChooserWindow::PCC2GameChooserWindow(Ptr<Directory> dir, bool oneup, PCC2OptionParser& parser)
    : UIWindow(_("Choose Game"), UIWindow::BLUE_WINDOW, COLOR_GRAY),
      dir_chooser(dir, oneup, 10),
      game_chooser(11, dir_chooser),
      parser(parser)
{
    setLayout(UIVBoxLayout::instance5);

    UIGroup& v      = holder.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& v2     = holder.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& h      = holder.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& h2     = holder.add(new UIGroup(UIHBoxLayout::instance5));
    UIButton& cd    = holder.add(new UIButton(0, _("Change directory"), 0, UIButton::aNothing));
    UIButton& watch = holder.add(new UIButton(0, _("Open Game"), 0, UIButton::aNothing));
    // Borland sucks.
    void (WGamedirChooser::*tmp)(int) = &WGamedirChooser::enter;
    cd.sig_fired.add(&dir_chooser, tmp);

    watch.sig_fired.add(this, &PCC2GameChooserWindow::maybeStop);
    game_chooser.sig_dblclick.add(this, &PCC2GameChooserWindow::maybeStop);

    h.add(cd);
    h.add(holder.add(new UISpacer()));
    h.add(watch);
    h.add(holder.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE, UIButton::aStop)));

    v.add(holder.add(new UIStaticText(_("Race:"), UIStaticText::BlackTitle)));
    v.add(game_chooser);
    v.add(holder.add(new UISpacer()));

    v2.add(holder.add(new UIStaticText(_("Game directory:"), UIStaticText::BlackTitle)));
    v2.add(dir_chooser);

    h2.add(v2);
    h2.add(v);

    add(h2);
    add(h);

    pack();
}

PCC2GameChooserWindow::~PCC2GameChooserWindow()
{ }

/** Switch to widget. Switches to widget by Id, if possible. */
void
PCC2GameChooserWindow::switchTo(int id)
{
    UIWidget* w = getWidgetById(id);
    if (w && !w->hasState(st_Disabled))
        w->setFocus();
}

/** Event handler. Handles Tab/Left/Right and Enter keys. */
bool
PCC2GameChooserWindow::handleEvent(const UIEvent& e, bool second_pass)
{
    UIWidget* w = getFocusedWidget();
    if (e.type == ev_Key) {
        switch (e.key.code) {
         case SDLK_TAB:
         case ss_Shift + SDLK_TAB:
            if (w && w->getId() == 10)
                switchTo(11);
            else
                switchTo(10);
            return true;
         case SDLK_RIGHT:
            switchTo(11);
            return true;
         case SDLK_LEFT:
            switchTo(10);
            return true;
         case SDLK_RETURN:
            if (w && w->getId() == 10) {
                dir_chooser.enter();
            } else {
                maybeStop(0);
            }
            return true;
         case vk_Quit:
            stop(cm_Escape);
            putKey(vk_Quit);
            return true;
        }
    }
    return UIWindow::handleEvent(e, second_pass);
}

/** Handler for "View VCRs" button. If possible, set parameters as
    entered in the dialog and stop() it. */
void
PCC2GameChooserWindow::maybeStop(int)
{
    if (Ptr<Directory> d = dir_chooser.getSubdirectory()) {
        parser.setGameDirName(d->getName());
        parser.setPlayerId(game_chooser.getCurrent()+1);
        stop(cm_Enter);
    }
}

int main(int /*argc*/, char** argv)
{
    Platform::initOutput();
    initNLS();
    PCC2OptionParser parser;
    Platform::initOptions(parser);
    initGraphicsOptions(parser);
    parser.parse(argv);

    SDL_Surface* sfc = initGraphics(0 /* FIXME: icon name */,
                                    "Planets Command Center II (v" PCC2_VERSION ")");
    if (!sfc) {
        Platform::showOutput();
        return 1;
    }

    try {
        initDebug();
        loadConfig();
        initUIGlobals(new GfxSurface(sfc, false), Directory::openDir(Platform::getResourceDirectory()));

        if (parser.getNumNonoptions() == 0) {
            initDirectories(Platform::getDataDirectory(), Platform::getDataDirectory());
            initRaceNames();
            parser.chooseGame(Directory::openDir(Directory::getWorkingDirectory()));
        }

        initGameGlobals(parser.getGameDirName(), parser.getRootDirName());
        initConfig();
        try {
            initRegInfo();
        }
        catch (FileOpenException& e) {
            console.write(LOG_WARN, _("No registration found, assuming unregistered player"));
        }
        initSpecification();
        initFCodes();
        initMissions();
        initRaceNames();
        initStarchart();

        GDirectoryOverview ov;
        ov.scan(*game_file_dir);

        console.write(LOG_INFO, _("=== Loading... ==="));
        if (int pid = parser.getPlayerId()) {
            if (! (ov.getPlayerFlags(pid) & (GDirectoryOverview::have_Gamedir | GDirectoryOverview::have_RST))) {
                console.write(LOG_ERROR, format(_("No data for player %d available"), pid));
                Platform::showOutput();
            } else {
                Ptr<Stream> pxy = spec_file_dir->openFileT("xyplan.dat", Stream::C_READ);
                getCurrentTurn().init();
                getCurrentTurn().loadPlanetXY(*pxy);
                if (ov.getPlayerFlags(pid) & GDirectoryOverview::have_RST) {
                    Ptr<Stream> rst =
                        game_file_dir->openFileT(format("player%d.rst", parser.getPlayerId()),
                                                 Stream::C_READ);
                    loadResult(getCurrentTurn(), *rst, parser.getPlayerId());
                    if (ov.getPlayerFlags(pid) & GDirectoryOverview::have_TRN) {
                        Ptr<Stream> trn =
                            game_file_dir->openFileT(format("player%d.trn", parser.getPlayerId()),
                                                     Stream::C_READ);
                        loadTurn(getCurrentTurn(), *trn, parser.getPlayerId());
                    }
                } else {
                    ASSERT(ov.getPlayerFlags(pid) & GDirectoryOverview::have_Gamedir);
                    loadDirectory(getCurrentTurn(), *game_file_dir, pid);
                }
                setPlayerIds(pid, pid);
                getCurrentTurn().postprocess();
                doPlayerScreen();

                // saving the game
                console.write(LOG_INFO, _("=== Saving... ==="));
                if (! (ov.getPlayerFlags(pid) & GDirectoryOverview::have_RST))
                    saveDirectory(getCurrentTurn(), *game_file_dir, GPlayerSet(pid));
                else
                    saveTurns(getCurrentTurn(), *game_file_dir, GPlayerSet(pid));
            }
        }
        console.write(LOG_INFO, _("Good bye."));
    }
    catch(std::exception& e) {
        console.write(LOG_ERROR, string_t(_("Error during initialisation: ")) + e.what());
        Platform::showOutput();
    }
    return 0;
}
