/**
  *  \file client/scr-env.cc
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens.h"
#include "util/nls.h"
#include "ui/layout.h"
#include "ui/window.h"
#include "ui/widgets.h"
#include "ui/uiglobal.h"
#include "gfx/palette.h"
#include "ui/util.h"
#include "game/planet.h"
#include "cpluslib/format.h"
#include "ui/msgbox.h"
#include "util/misc.h"

/*
      Neutronium
      abundand, large masses
      10000kt [######-------]
      10000kt [######.......]
      99%     [###..........]

  */

WPlanetMiningTile::WPlanetMiningTile(GPlanet& planet, GCargoType ele)
    : WTile(295, 4 * font_heights[FONT_NORMAL] + font_heights[FONT_TITLE] + 5),
      planet(planet), ele(ele)
{ }

void
WPlanetMiningTile::drawFixed(GfxCanvas& can)
{ }

static void
drawPercentBar(GfxCanvas& can, GfxRect rect, long fill, int divi, int color)
{
    int split = rect.w * fill / (100*divi);
    if (fill && !split)
        split = 1;
    if (split > rect.w)
        split = rect.w;

    drawSolidBar(can, GfxRect(rect.x, rect.y, split, rect.h), standard_colors[color]);
    drawSolidBar(can, GfxRect(rect.x+split, rect.y, rect.w-split, rect.h), standard_colors[COLOR_SHIELD+3]);
}

static const char*
massToString(int32 mass)
{
    if (mass >= 5000)
        return _("abundand");
    else if (mass >= 1200)
        return _("very common");
    else if (mass >= 600)
        return _("common");
    else if (mass >= 100)
        return _("rare");
    else
        return _("very rare");
}

static const char*
densityToString(int dens)
{
    if (dens < 10)
        return _("very scattered");
    else if (dens < 30)
        return _("scattered");
    else if (dens < 40)
        return _("dispersed");
    else if (dens < 70)
        return _("concentrated");
    else
        return _("large masses");
}

void
WPlanetMiningTile::drawData(GfxCanvas& can)
{
    int y = 3;
    
    drawFrameDown(can, getExtent());

    setColor(UIColor::tc_White);
    outTextF(can, FONT_NORMAL, 5, y, extent.w-10, getCargoName(ele));
    y += font_heights[FONT_NORMAL] + 3;

    getSkin().setSolid(true);
    setColor(UIColor::tc_Static);
    if (planet.getCargo(ele))
        outTextF(can, FONT_NORMAL, 5, y, extent.w-10,
                 format("%s, %s", massToString(planet.getOreGround(ele)),
                        densityToString(planet.getOreDensity(ele))));
    else
        outTextF(can, FONT_NORMAL, 5, y, extent.w-10, _("none"));
    y += font_heights[FONT_NORMAL];

    outText(can, FONT_NORMAL, 5, y, _("mined"));
    drawPercentBar(can, GfxRect(extent.x+60, extent.y+y, 150, 15),
                   planet.getCargo(ele), 82, COLOR_WHITE);
    getSkin().getParameters().just_x = 2;
    outTextF(can, FONT_NORMAL, extent.w - 25, y, 60, numToString(planet.getCargo(ele)));
    getSkin().getParameters().just_x = 0;
    outTextF(can, FONT_NORMAL, extent.w - 25, y, 20, " kt");
    y += font_heights[FONT_NORMAL];

    outText(can, FONT_NORMAL, 5, y, _("ground"));
    drawPercentBar(can, GfxRect(extent.x+60, extent.y+y, 150, 15),
                   planet.getOreGround(ele), 82, COLOR_WHITE);
    getSkin().getParameters().just_x = 2;
    outTextF(can, FONT_NORMAL, extent.w - 25, y, 60, numToString(planet.getOreGround(ele)));
    getSkin().getParameters().just_x = 0;
    outTextF(can, FONT_NORMAL, extent.w - 25, y, 20, " kt");
    y += font_heights[FONT_NORMAL];

    outText(can, FONT_NORMAL, 5, y, _("density"));
    drawPercentBar(can, GfxRect(extent.x+60, extent.y+y, 150, 15),
                   planet.getOreDensity(ele), 1, COLOR_GRAY);
    getSkin().getParameters().just_x = 2;
    outTextF(can, FONT_NORMAL, extent.w - 25, y, 60, numToString(planet.getOreDensity(ele)));
    getSkin().getParameters().just_x = 0;
    outTextF(can, FONT_NORMAL, extent.w - 25, y, 20, " %");
    y += font_heights[FONT_NORMAL];
}

namespace {
    class WPlanetInfoBox : public UIWidget {
        GPlanet& planet;
        int draw_y;
        int font;
        void write(GfxCanvas& can, int x, string_t text);
        void showSupport(GfxCanvas& can, int player);
     public:
        WPlanetInfoBox(GPlanet& planet);
        void draw(GfxCanvas& can, UpdateInfo& ui);
        void getLayoutInfo(LayoutInfo& li);
    };
}

WPlanetInfoBox::WPlanetInfoBox(GPlanet& planet)
    : planet(planet)
{ }

void
WPlanetInfoBox::write(GfxCanvas& can, int x, string_t text)
{
    getSkin().outTextF(can, font, extent.x + x, extent.y + draw_y, extent.w - x, text);
    draw_y += font_heights[font];
}

void
WPlanetInfoBox::showSupport(GfxCanvas& can, int player)
{
    write(can, 5, format(_("Supports %s %ss"),
                         clansToString(planet.getMaxSupportedColonists()),
                         getAdjName(player)));
    // FIXME 'won't die ...'
}

void
WPlanetInfoBox::draw(GfxCanvas& can, UpdateInfo& ui)
{
    getSkin().startDraw();

    draw_y = 0;
    font = FONT_TITLE;
    getSkin().setColor(UIColor::tc_White);
    getSkin().setSolid(true);
    write(can, 0, _("Climate"));
    font = FONT_NORMAL;
    getSkin().setColor(UIColor::tc_Static);

    write(can, 5, format(_("Climate type: %s"), planet.getTempStr()));
    write(can, 5, format(_("Average temperature: %dF"), planet.getTemp()));
    if (planet.getOwner() > 0 && planet.getOwner() != ::getPlayerId())
        showSupport(can, planet.getOwner());
    showSupport(can, ::getPlayerId());

    draw_y += 10;
    font = FONT_TITLE;
    getSkin().setColor(UIColor::tc_White);
    write(can, 0, _("Population"));
    font = FONT_NORMAL;
    getSkin().setColor(UIColor::tc_Static);

    if (planet.getNativeRace() == 0) {
        write(can, 5, _("No native population"));
    } else {
        write(can, 5, format(_("Native race: %s"),
                             planet.getNativeRaceName(planet.getNativeRace())));
        write(can, 5, format(_("Population: %s"), clansToString(planet.getNatives())));
        write(can, 5, format(_("Government: %s"),
                             planet.getNativeGovernmentName(planet.getNativeGovernment())));
    }
    // FIXME: assimilation forecast

    draw_y += 3;
    if (planet.getOwner() == 0) {
        write(can, 5, _("No colonists"));
    } else {
        write(can, 5, format(_("Colonists: %s"), getAdjName(planet.getOwner())));
        write(can, 5, format(_("Population: %s"), clansToString(planet.getColonists())));

        string_t industry;
        if (planet.getStructures(str_Factories))
            industry += format(_(", %d factor%!1{ies%|y%}"), planet.getStructures(str_Factories));
        if (planet.getStructures(str_Mines))
            industry += format(_(", %d mine%!1{s%}"), planet.getStructures(str_Mines));
        if (planet.getStructures(str_Defense))
            industry += format(_(", %d DP%!1{s%}"), planet.getStructures(str_Defense));
        if (!industry.empty())
            write(can, 5, industry.substr(2));

        write(can, 5, format(_("%d mc, %d suppl%!1{ies%|y%}"),
                             planet.getCargo(el_Money),
                             planet.getCargo(el_Supplies)));
        write(can, 5, format(_("Friendly code: %s"),
                             planet.getFCode()));
        if (planet.hasBase())
            write(can, 5, _("Starbase present"));
        else if (planet.isBuildingBase())
            write(can, 5, _("Starbase being built"));
    }
    // FIXME: unload amount

    ui.update(getExtent());
}

/* Display on info box:
    <big> Climate
    Climate type: Temp               or: (no information available)
    Average temperature: 99F
    <support for owner>
    <support for us>

    <big>Population
    no natives
      or: Native race: foo
          Population:  100c
          Government:  bar
          Assimilated in...
          (age)
    no information on colonists
      or: Colonists: race
          Population: ...
          ... industrial activity
          or: n factories, n defense, n mines
          (age)
          n mc, n supplies
          (age)
          Friendly code: ...
          or: Last known fcode: ...
          Starbase present/being built
          beaming down NN clans, imperial assault */

void
WPlanetInfoBox::getLayoutInfo(LayoutInfo& li)
{
    li.flags = 0;
    li.pref_size = li.min_size = GfxPoint(310, 25 * font_heights[FONT_NORMAL]);
}

int
doPlanetScan(int pid)
{
    GPlanet& p = getDisplayedTurn().getPlanet(pid);
    if (!p.haveDat()) {
        messageBox(format(_("No information available about planet %s (#%d)."),
                          p.getName(), pid),
                   _("Planet Scan"));
        return 0;
    }

    UIWindow win(p.getName(), UIWindow::BLUE_WINDOW, COLOR_DARK);
    win.setLayout(UIHBoxLayout::instance5);
    UIGroup lgroup(UIVBoxLayout::instance5), rgroup(UIVBoxLayout::instance5),
        bgroup(UIHBoxLayout::instance5);
    WPlanetMiningTile n(p, el_Neutronium),
        t(p, el_Tritanium),
        d(p, el_Duranium),
        m(p, el_Molybdenum);
    UISpacer sp1;
    WPlanetInfoBox info(p);
    UIButton ok(1, _("Close"), SDLK_RETURN, UIButton::aStop);

    bgroup.add(ok);
    bgroup.add(sp1);
    lgroup.add(info);
    lgroup.add(bgroup);
    rgroup.add(n);
    rgroup.add(t);
    rgroup.add(d);
    rgroup.add(m);
    win.add(lgroup);
    win.add(rgroup);
    win.pack();
    ui_root->centerWidget(win);
    ui_root->add(win);
    ui_root->run();
    ui_root->close(win);
    return 0;
}
