/**
  *  \file game/act-build.cc
  *  \brief Ship Building Action
  *
  *  This file contains the GStarbaseBuildShipAction action which
  *  encapsulates the logic for building a ship.
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/action.h"
#include "game/spec.h"
#include "game/player.h"
#include "util/nls.h"

/** Ship Build Action. */

GStarbaseBuildShipAction::GStarbaseBuildShipAction(GPlanet& planet, GPlanetTransfer& pay)
    : GStarbaseBuildAction(planet, pay),
      order(planet.getBaseBuildOrder()),
      use_parts_from_storage(true),
      use_tech_upgrade(true),
      is_ok(true)
{
    if (!order.hull) {
        // no pending build order. Invent one.
        order.hull = getTruehull(getPlanet().getOwner(), 1);
        order.beam_count = getHull(order.hull).getMaxBeams();
        order.torp_count = getHull(order.hull).getMaxLaunchers();
        order.engine = 1;
        order.beam_type = 1;
        order.torp_type = 1;
        while (order.engine < NUM_ENGINES && getEngine(order.engine+1).getTech() <= planet.getBaseTech(tl_Engine))
            ++order.engine;
        while (order.beam_type < NUM_BEAMS && getBeam(order.beam_type+1).getTech() <= planet.getBaseTech(tl_Beam))
            ++order.beam_type;
        while (order.torp_type < NUM_TORPS && getTorp(order.torp_type+1).getTech() <= planet.getBaseTech(tl_Torpedo))
            ++order.torp_type;
        use_parts_from_storage = false;
    } else {
        if (!order.beam_count) {
            order.beam_type = 1;
            while (order.beam_type < NUM_BEAMS && getBeam(order.beam_type+1).getTech() <= planet.getBaseTech(tl_Beam))
                ++order.beam_type;
        }
        if (!order.torp_count) {
            order.torp_type = 1;
            while (order.torp_type < NUM_TORPS && getTorp(order.torp_type+1).getTech() <= planet.getBaseTech(tl_Torpedo))
                ++order.torp_type;
        }
    }
    regenerate();
}

GStarbaseBuildShipAction::~GStarbaseBuildShipAction()
{ }

/** Choose whether parts from storage will be used. If enabled, ship
    building will use parts if possible. If disabled, everything will
    be built anew, even when there is already a matching part in
    storage. */
void
GStarbaseBuildShipAction::setUsePartsFromStorage(bool b)
{
    if (b != use_parts_from_storage) {
        use_parts_from_storage = b;
        regenerate();
    }
}

/** Choose whether tech level upgrades are allowed. If not, we'll
    refuse to build any ship that needs an upgrade. */
void
GStarbaseBuildShipAction::setUseTechUpgrade(bool b)
{
    if (b != use_tech_upgrade) {
        use_tech_upgrade = b;
        regenerate();
    }
}

GError
GStarbaseBuildShipAction::getError()
{
    if (!is_ok)
        return GError(GError::ePerm, _("This build order cannot be performed"));
    return GStarbaseBuildAction::getError();
}

void
GStarbaseBuildShipAction::setBuildOrder(const TBuildOrder& o)
{
    order = o;
    regenerate();
}

/** Do a tech upgrade. This accounts resources, and sets is_ok to
    false if anything doesn't want us to upgrade.
    \param tl     which level we're talking about
    \param a      actor whose setTech we ought to call
    \param need   required level */
void
GStarbaseBuildShipAction::doTechUpgrade(TTechLevel tl, Actor& a, int need)
{
    /* do we need an upgrade? */
    if (getPlanet().getBaseTech(tl) >= need)
        return;

    /* are we allowed to upgrade? */
    if (need > 6 && !allowRegistered(getPlanet().getOwner()))
        is_ok = false;

    /* we need a tech upgrade but do not want to. In particular, do
       not include upgrades in the calculation. */
    if (!use_tech_upgrade) {
        is_ok = false;
        return;
    }

    a.setTech(tl, need);
}

/** Compute how many parts we have to build.
    \param need   we need this many parts...
    \param have   ...and we have this many already in storage */
inline int
GStarbaseBuildShipAction::getBuildAmount(int need, int have)
{
    if (!use_parts_from_storage)
        return need;
    else if (need <= have)
        return 0;
    else
        return need - have;
}

/** Gegenerate all orders, and update validity status. We have some
    more validity checks than just 'sufficient funds':
    - can we upgrade tech?
    - did the user choose no tech upgrade but we would need one? */
void
GStarbaseBuildShipAction::doChanges(Actor& a)
{
    is_ok = true;

    // first, attempt to build one hull
    if (int slot = getTruehullSlot(getPlanet().getOwner(), order.hull)) {
        // we can build it
        if (int n = getBuildAmount(1, getPlanet().getBaseHullStoreSlot(slot))) {
            doTechUpgrade(tl_Hull, a, getHull(order.hull).getTech());
            a.setHullSlot(slot, getPlanet().getBaseHullStoreSlot(slot) + n);
        }
    } else {
        // we can not build it.
        // FIXME: we should anyway add the cost somehow
        is_ok = false;
        doTechUpgrade(tl_Hull, a, getHull(order.hull).getTech());
    }

    // build the engines
    if (int n = getBuildAmount(getHull(order.hull).getNumEngines(),
                               getPlanet().getBaseEngineStore(order.engine)))
    {
        doTechUpgrade(tl_Engine, a, getEngine(order.engine).getTech());
        a.setEngine(order.engine, getPlanet().getBaseEngineStore(order.engine) + n);
    }

    // build the arms
    int n;
    
    if (order.beam_count && (n = getBuildAmount(order.beam_count, getPlanet().getBaseBeamStore(order.beam_type))) != 0) {
        doTechUpgrade(tl_Beam, a, getBeam(order.beam_type).getTech());
        a.setBeam(order.beam_type, getPlanet().getBaseBeamStore(order.beam_type) + n);
    }

    if (order.torp_count && (n = getBuildAmount(order.torp_count, getPlanet().getBaseLauncherStore(order.torp_type))) != 0) {
        doTechUpgrade(tl_Torpedo, a, getTorp(order.torp_type).getTech());
        a.setBeam(order.torp_type, getPlanet().getBaseLauncherStore(order.torp_type) + n);
    }
}

void
GStarbaseBuildShipAction::commit()
{
    GStarbaseBuildAction::commit();

    TBuildOrder o = order;
    if (!o.torp_count)
        o.torp_type = 0;
    if (!o.beam_count)
        o.beam_type = 0;
    o.unused = 0;

    getPlanet().setBaseBuildOrder(o);
}
