/**
  *  \file game/cargo-tr.cc
  *  \brief Cargo Transfer and Cargo Cost Transaction
  *
  *  This module implements the two most basic cargo transactions.
  *  GCargoTransfer is a means of moving cargo between objects.
  *  A GCargoCostTransaction is a transaction which turns cargo into
  *  something else.
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "game/cargo.h"
#include "cpluslib/assert.h"
#include "util/nls.h"

/***************************** GCargoTransfer ****************************/

/*! \class GCargoTransfer

    This class provides everything necessary to transfer cargo
    between two objects. */

/** Constructor.
    \param left left object
    \param right right object */
GCargoTransfer::GCargoTransfer(GCargoContainer& left, GCargoContainer& right)
{
    objs[0] = &left;
    objs[1] = &right;
}

/** Destructor. Cancels the transaction if it was not yet finished. */
GCargoTransfer::~GCargoTransfer()
{
    if (!isFinished())
        cancel();
}

/** Move stuff.
    \param type    type to move
    \param amount  amount to move. May be negative; semantics are
                   preserved.
    \param from    source object (0=left, 1=right)
    \param partial true to allow partial transfers, false to reject
    \param s2m     sell supplies on-the-fly
    \return amount moved.
    \inv !partial => return == 0 || return == amount
    \inv partial  => return >= 0 && return <= amount */
int32
GCargoTransfer::move(GCargoType type, int32 amount, int from, bool partial,
                     bool s2m)
{
    /* trivial case? */
    if (!amount)
        return 0;

    /* reverse transfer? */
    if (amount < 0)
        return -move(type, -amount, 1-from, partial, false);

    /* figure out src/dest objects and type */
    GCargoContainer& src  = *objs[from];
    GCargoContainer& dest = *objs[1 - from];

    GCargoType dtype = (type == el_Supplies && s2m ? el_Money : type);

    /* can we actually move it? */
    if (!src.canHave(type) || !dest.canHave(dtype))
        return 0;

    int32 space = dest.getMax(dtype) - dest.getCurrent(dtype);
    if (amount > space) {
        if (!partial)
            return 0;
        amount = space;
    }

    /* do the transfer */
    int32 got = src.removeCargo(type, amount, partial);
    int32 put = dest.addCargo(dtype, got, partial);
    ASSERT(got == put);
    ASSERT(got <= amount);

    if (got)
        sig_changed.raise();

    return got;
}

void
GCargoTransfer::cancel()
{
    GTransaction::cancel();
    objs[0]->mustCancel();
    objs[1]->mustCancel();
}

void
GCargoTransfer::commit()
{
    GTransaction::commit();
    objs[0]->mustCommit();
    objs[1]->mustCommit();
}

/************************* GCargoCostTransaction *************************/

/** \class GCargoCostTransaction
    \brief General Cost Transaction

    This is the base class for all transactions which cost some
    resources, for example, buying structures, building a base or
    raising tech. It provides functions to change the bill, and do the
    allocation on a cargo container.

    This class supports costs consisting of the three industry
    minerals (T/D/M), money and supplies.

    The usual failure mode for a GCargoCostTransaction is lacking
    material in the cargo container. In addition to that failure mode,
    we support managing a GError object. */

/** Constructor. Initialize with cost zero. Note: the container should
    be a planet; we assume that addCargo never fails. */
GCargoCostTransaction::GCargoCostTransaction(GCargoContainer& container)
    : container(container)
{
    for (int i = 0; i < LIMIT; ++i)
        cost_amounts[i] = reserved_amounts[i] = 0;
}

/** Destructor. Cancels the transaction if not already finished. */
GCargoCostTransaction::~GCargoCostTransaction()
{
    if (!isFinished())
        cancel();
}

/** Validate cargo type. We only support a subset of types. */
inline bool
GCargoCostTransaction::isValidIndex(GCargoType t) const
{
    return (t == el_Tritanium || t == el_Duranium || t == el_Molybdenum
            || t == el_Money || t == el_Supplies);
}

/** Check if transaction is valid. The transaction can only be
    committed when this returns true.
    \inv isValid() <=> getMissing(T) == 0 for all valid Ts */
bool
GCargoCostTransaction::isValidCost() const
{
    /* FIXME: not yet checked whether this is safe for giving back
       stuff */
    return reserved_amounts[el_Tritanium]  >= cost_amounts[el_Tritanium]
        && reserved_amounts[el_Duranium]   >= cost_amounts[el_Duranium]
        && reserved_amounts[el_Molybdenum] >= cost_amounts[el_Molybdenum]
        && reserved_amounts[el_Supplies]   >= cost_amounts[el_Supplies]
        && (reserved_amounts[el_Supplies] + reserved_amounts[el_Money]
            >= cost_amounts[el_Supplies] + cost_amounts[el_Money]);
}

/** Change the cost.
    \param type type of resource to modify
    \param amount amount to add (negative to subtract)
    \see setCost() */
void
GCargoCostTransaction::changeCost(GCargoType type, int32 amount)
{
    setCost(type, cost_amounts[type] + amount);
}

/** Set cost. This does actual re-allocation.
    \post getCost(type) == amount */
void
GCargoCostTransaction::setCost(GCargoType type, int32 amount)
{
    ASSERT(isValidIndex(type));

    const bool raise = (cost_amounts[type] != amount);

    cost_amounts[type] = amount;

    /* if it's a mineral, it's simple */
    if (type == el_Tritanium || type == el_Duranium || type == el_Molybdenum) {
        if (amount > reserved_amounts[type]) {
            /* we want more, allocate some */
            reserved_amounts[type] +=
                container.removeCargo(type, amount - reserved_amounts[type], true);
        } else if (amount < reserved_amounts[type]) {
            /* we want less, give something back */
            reserved_amounts[type] -=
                container.addCargo(type, reserved_amounts[type] - amount, true);
        }
    } else {
        /* money or supplies. First give back everything, then
           re-acquire it. */
        container.addOrRemoveCargo(el_Money,    reserved_amounts[el_Money],    true);
        container.addOrRemoveCargo(el_Supplies, reserved_amounts[el_Supplies], true);

        reserved_amounts[el_Money] =
            -container.addOrRemoveCargo(el_Money, -cost_amounts[el_Money], true);
        reserved_amounts[el_Supplies] =
            -container.addOrRemoveCargo(el_Supplies,
                                        -(cost_amounts[el_Supplies]
                                          + cost_amounts[el_Money] - reserved_amounts[el_Money]),
                                        true);
    }
    if (raise)
        sig_changed.raise();
}

/** Get cost. */
int32
GCargoCostTransaction::getCost(GCargoType type) const
{
    ASSERT(isValidIndex(type));
    return cost_amounts[type];
}

/** Get remaining stuff after this transaction. */
int32
GCargoCostTransaction::getRemaining(GCargoType type) const
{
    ASSERT(isValidIndex(type));
    return container.getCurrent(type);
}

/** Get missing stuff for this transaction. When there's something
    missing, isValid() will be false.
    \inv isValid() => return value will be 0 */
int32
GCargoCostTransaction::getMissing(GCargoType type) const
{
    ASSERT(isValidIndex(type));

    if (type == el_Supplies) {
        /* missing supplies. We're lacking supplies only when the cost
           actually includes some. */
        if (cost_amounts[el_Supplies] <= reserved_amounts[el_Supplies])
            return 0;
        else
            return cost_amounts[el_Supplies] - reserved_amounts[el_Supplies];
    } else if (type == el_Money) {
        /* missing money. When we have too little money but can
           compensate with supplies, pretend we're not lacking cash. */
        int32 unused_supplies = reserved_amounts[el_Supplies] - cost_amounts[el_Supplies];
        if (unused_supplies < 0)
            unused_supplies = 0;
        int32 missing_money = cost_amounts[el_Money] - reserved_amounts[el_Money] - unused_supplies;
        if (missing_money < 0)
            return 0;           // FIXME: can this happen?
        else
            return missing_money;
    } else {
        /* minerals don't convert */
        return cost_amounts[type] - reserved_amounts[type];
    }
}

/** Get reserved amounts of a given type. This value is equal to the
    actual cost of the transaction, but in case of supplies sold for
    money, it yields the actual amounts used. */
int32
GCargoCostTransaction::getReserved(GCargoType type) const
{
    ASSERT(isValidIndex(type));
    return reserved_amounts[type];
}

/** Commit transaction. Caller must commit the GCargoContainer to actually
    use the minerals (and do the thing this cost is for, of course). */
void
GCargoCostTransaction::commit()
{
    ASSERT(isValidCost());
    container.mustCommit();
    GTransaction::commit();
}

/** Clear costs. This sets all costs to zero. */
void
GCargoCostTransaction::clearCosts()
{
    setCost(el_Tritanium, 0);
    setCost(el_Duranium, 0);
    setCost(el_Molybdenum, 0);
    setCost(el_Supplies, 0);
    setCost(el_Money, 0);
}

/** Cancel transaction. Caller must cancel the GCargoContainer, too. */
void
GCargoCostTransaction::cancel()
{
    clearCosts();
    container.mustCancel();
    GTransaction::cancel();
}

/** Get error message. */
GError
GCargoCostTransaction::getError()
{
    if (!isValidCost())
        return GError(GError::eNoResource, _("Not enough resources to perform this action"));
    return GError();
}
