/**
  *  \file game/cargo.cc
  *  \brief Cargo Reservation/Arbitration Primitives
  *
  *  PCC II supports "concurrent" access to cargo containers. Hence,
  *  we use a transaction-like scheme to ensure correct behaviour. 
  *  Each object has a cargo arbiter (GCargo), and a number of cargo
  *  traits (GCargoContainer). Cargo which is removed from the object
  *  is first reserved, and not added to the target object until the
  *  transaction is committed. We don't support "real", preemptive
  *  concurrency, though.
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "game/cargo.h"
#include "cpluslib/assert.h"
#include "cpluslib/misc.h"
#include "util/nls.h"
#include "game/spec.h"

/********************************* GCargo ********************************/

/** Constructor. */
GCargo::Impl::Impl()
    : ref_count(0)
{
    std::fill_n(reserved_cargo, int(el_COUNT), 0);
}

GCargo::~GCargo()
{
    delete data;
}

/** Increment use count. Every piece of code which possibly does a
    cargo reservation must call this. */
void
GCargo::incUseCount()
{
    if (!data)
        data = new Impl();
    ++data->ref_count;
}

/** Decrement use count. You may no longer call changeReservation()
    after calling decUseCount(). */
void
GCargo::decUseCount() throw()
{
    if (!--data->ref_count) {
        delete data;
        data = 0;
    }
}

/** Change cargo reservation.
    \param t cargo type
    \param amount >0 to reserve cargo, <0 to give back
    \returns new amount of reserved cargo. */
int32
GCargo::changeReservation(GCargoType t, int32 amount) throw()
{
    ASSERT(data);
    ASSERT(data->reserved_cargo[t] + amount >= 0);
    data->reserved_cargo[t] += amount;
    return data->reserved_cargo[t];
}

/** Acquire facility n. Only one process can use a facility at a given
    time.
    \return true iff facility successfully acquired
    \see releaseFacility */
bool
GCargo::acquireFacility(int n) throw()
{
    if (isFacilityInUse(n))
        return false;
    used_facilities |= 1L << n;
    return true;
}

/** Release facility n. */
void
GCargo::releaseFacility(int n) throw()
{
    ASSERT(isFacilityInUse(n));
    used_facilities &= ~(1L << n);
}

/**************************** GCargoContainer ****************************/

/*! \class GCargoContainer

    This is the base class for everything that deals with cargo
    (minerals, sup, mc, fighters, torps, clans). It is used for cargo
    transfer and some other things.

    Derived classes provide access to the underlying object.
    - getAmount() yields the current amount on the object, where
      reserved stuff has already been subtracted (that is, the
      amount which can still be safely subtracted).
    - changeAmount() changes the current physical amount.
    - reserveAmount() changes the current reserved amount.
    - getMax() yields the maximum amount which can appear on the
      object.

    The public manipulation interface uses these functions to provide
    safe cargo amount manipulation. Using commit(), you can commit the
    changes; cancel() takes them back.

    \todo FIXME: a problem which can not be solved is to test whether a
    package of multiple cargo types can fit onto a container without
    modifying it, e.g. to test whether I can add "50TDM 2800$". Things
    I want to do with this:
    - "load minerals for as many torps as possible" (mkt support)
    - clearing the hold space in multi-cargo transfer.

    It is still possibly doing this item-by-item, but that is
    cumbersome.

    Interfaces that come to mind (public virtual functions):
    - bool canAdd(GCargoContainer& other): return true iff other
      can be added to *this. This solves problem 2; problem 1 can
      be solved using binary intersection tests.
    - int divide(GCargoContainer& other): how many times other
      can be added to *this. Elegant at first, this one has problems
      handling "infinite" containers (planets). */

/** Constructor. Creates a cargo container which did not yet
    add/remove anything from the underlying object. */
GCargoContainer::GCargoContainer()
{
    std::fill_n(diffs, countof(diffs), 0);
    std::fill_n(reserved, countof(reserved), 0);
}

GCargoContainer::~GCargoContainer()
{ }

/** Return current amount of cargo. This is the value to be presented
    to the user. It includes others' removals but not their additions,
    and it includes our removals and additions. Therefore it is safe
    in the sense that the value returned from this function can be
    removed from the object without problems. */
int32
GCargoContainer::getCurrent(GCargoType type)
{
    if (diffs[type] < 0) {
        /* we have removed some cargo from this object. Hence, getAmount()
           will include our reservation. */
        ASSERT(reserved[type] == -diffs[type]);
        return getAmount(type);
    } else {
        /* we are receiving cargo. getAmount() will include others'
           reservations. */
        ASSERT(reserved[type] == 0);
        return getAmount(type) + diffs[type];
    }
}

/** Remove some cargo from this object. This will not actually remove
    the cargo from the underlying object, it will merely reserve it
    for us.
    \param type     cargo type
    \param amount   amount to remove
    \param partial  allow partial removal

    \return amount removed. If partial=false, either 0 or /amount/. 
    Otherwise, any value from [0, amount]. */
int32
GCargoContainer::removeCargo(GCargoType type, int32 amount, bool partial)
{
    ASSERT(amount >= 0);

    int32 have = getCurrent(type);
    if (amount > have) {
        if (!partial)
            return 0;
        amount = have;
    }

    if (diffs[type] > 0) {
        if (diffs[type] >= amount) {
            diffs[type] -= amount;
            return amount;
        } else {
            amount -= diffs[type];
            diffs[type] = 0;
        }
    }
    reserveAmount(type, amount);
    reserved[type] += amount;
    diffs[type] -= amount;
    if (amount)
        sig_changed.raise();
    return amount;
}

/** Add cargo to this object. This will not actually add the cargo to
    the object, but will just remember to do that at a later time.
    \param type    cargo type
    \param amount  amount to add
    \param partial allow partial add

    \return amount added. If partial=false, either 0 or /amount/. 
    Otherwise, any value from [0,amount]. */
int32
GCargoContainer::addCargo(GCargoType type, int32 amount, bool partial)
{
    ASSERT(amount >= 0);

    /* add some cargo. First check whether we can actually have that much. */
    int32 space = getMax(type) - getCurrent(type);
    if (amount > space) {
        if (!partial)
            return 0;
        amount = space;
    }

    /* can we add this by cancelling a pending "get"? */
    if (reserved[type]) {
        if (reserved[type] > amount) {
            reserveAmount(type, -amount);
            reserved[type] -= amount;
        } else {
            reserveAmount(type, -reserved[type]);
            reserved[type] = 0;
        }
    }
    diffs[type] += amount;
    if (amount)
        sig_changed.raise();
    return amount;
}

/** Add or remove cargo. Same as addCargo() if amount is positive,
    same as removeCargo() if amount is negative. See there for
    details. */
int32
GCargoContainer::addOrRemoveCargo(GCargoType type, int32 amount, bool partial)
{
    if (amount > 0)
        return addCargo(type, amount, partial);
    else if (amount < 0)
        return -removeCargo(type, -amount, partial);
    else
        return 0;
}

/** Cancel transaction. This will undo our reservations. */
void
GCargoContainer::cancel()
{
    GTransaction::cancel();

    /* cancelling a transaction means canceling our reservations */
    for (GCargoType i = el_FIRST; i < el_COUNT; ++i)
        reserveAmount(i, -reserved[i]);
}

/** Commit transaction. This will perform all changes. */
void
GCargoContainer::commit()
{
    GTransaction::commit();

    /* commit means to cancel our reservations *and* apply our changes */
    for (GCargoType i = el_FIRST; i < el_COUNT; ++i) {
        reserveAmount(i, -reserved[i]);
        ASSERT(getAmount(i) + diffs[i] >= 0);
        if (diffs[i])
            changeAmount(i, diffs[i]);
    }
}

/** Get user-visible name of cargo item /type/. */
string_t
getCargoName(GCargoType type)
{
    static const char* const names[el_Torps] = {
        N_("Neutronium"),
        N_("Tritanium"),
        N_("Duranium"),
        N_("Molybdenum"),
        N_("Fighters"),
        N_("Colonists"),
        N_("Supplies"),
        N_("Money")
    };
    if (type < el_Torps)
        return string_t(_(names[type]));
    else
        return ::getTorp(getTorpTypeFromCargoType(type)).getName();
}

/** Get unit of cargo type /type/.
    \returns name of unit, or empty string if type doesn't have
    a unit (i.e. torps) */
string_t
getCargoUnit(GCargoType type)
{
    switch(type) {
     case el_Colonists:
        return string_t(_("Clans"));
     case el_Neutronium:
     case el_Duranium:
     case el_Tritanium:
     case el_Molybdenum:
     case el_Supplies:
        return string_t("kt");  // I guess we better don't translate that
     case el_Money:
        return string_t("mc");  // same here
     default:
        return string_t();
    }
}
