/**
  *  \file game/cargo.h
  *  \brief Cargo Management
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_CARGO_H
#define PCC_V2_GAME_CARGO_H

#include "cpluslib/assert.h"
#include "cpluslib/types.h"
#include "game/struct.h"
#include "game/transact.h"
#include "game/error.h"

/** Cargo type. */
enum GCargoType {
    el_Neutronium,              ///< Neutronium.
    el_Tritanium,               ///< Tritanium.
    el_Duranium,                ///< Duranium.
    el_Molybdenum,              ///< Molybdenum.
    el_Fighters,                ///< Fighters.
    el_Colonists,               ///< Colonist clans.
    el_Supplies,                ///< Supplies.
    el_Money,                   ///< Cash.
    el_Torps,                   ///< Torps. \see getTorpTypeFromCargoType(), getCargoTypeFromTorpType()
    el_COUNT = el_Torps + NUM_TORPS,
    el_FIRST = el_Neutronium
};

/** Cargo Arbiter. This class arbitrates interleaved / overlapping
    cargo transaction.

    - when you manipulate cargo, obtain a cargo reservation object
      (foo.getCargoHandler().incUseCount())
    - when you wish to remove cargo from somewhere, first reserve
      the cargo on the cargo handler. It will disappear from display.
    - finally, you can commit the transaction by removing the
      reservation, and performing the pending add and remove operations.
    - drop the reservation object (foo.getCargoHandler().decUseCount()))

    In addition, every object can have several "facilities" (up to
    32, numbered 0 to 31) which, for simplicity, don't have their
    own cargo arbiter. Only one entity can use such a facility at a
    time. Facilities include:
    - a ship's cargo transporters
    - a starbase's hull/engine/beam/torp launcher store

    The cargo arbiter does not itself know how much cargo is
    available, nor does it check any rules. It just manages
    reservations and facilities. */
class GCargo {
 private:
    struct Impl {
        /** Reference count. Equals the number of incUseCount() calls
            which have not been taken back using decUseCount() yet. */
        int ref_count;
        /** Reserved cargo which is not available. */
        int32 reserved_cargo[el_COUNT];

        Impl();
    };
    Impl*   data;
    uint32  used_facilities;
 public:
    GCargo()
        : data(0), used_facilities(0) { }
    ~GCargo();

    class FacilityArbiter;

    void incUseCount();
    void decUseCount() throw();

    /** Return reserved amount, for user interface purposes. */
    int32 getReserved(GCargoType t) const
        { return data ? data->reserved_cargo[t] : 0; }

    /** Change reservation. */
    int32 changeReservation(GCargoType t, int32 diff) throw();

    /** Query whether facility is in use. */
    bool isFacilityInUse(int n) const
        { return (used_facilities & (1L << n)) != 0; }

    bool acquireFacility(int n) throw();
    void releaseFacility(int n) throw();
};

/** Facility Arbiter. This helper class makes acquiring and releasing
    a facility exception-safe. It also manages the use count of the
    GCargo object. */
class GCargo::FacilityArbiter {
    GCargo& cargo;
    short   facility;  // use short so this object fits into two words
    bool    have_it;
 public:
    /** Constructor. Acquire the facility and increment the use
        count. */
    FacilityArbiter(GCargo& cargo, int facility)
        : cargo(cargo), facility(facility), have_it(cargo.acquireFacility(facility))
        { cargo.incUseCount(); }
    /** Destructor. Release the facility and decrement the use
        count. */
    ~FacilityArbiter()
        {
            if (have_it)
                cargo.releaseFacility(facility);
            cargo.decUseCount();
        }
    /** Check whether we got the facility. */
    operator bool() const
        { return have_it; }
};

/* bummer. Why can't C++ have these built-in? */
inline GCargoType operator++(GCargoType& t) { return t = GCargoType(t+1); }
inline GCargoType operator--(GCargoType& t) { return t = GCargoType(t-1); }
inline GCargoType operator++(GCargoType& t,int) { GCargoType rv = t; ++t; return rv; }
inline GCargoType operator--(GCargoType& t,int) { GCargoType rv = t; --t; return rv; }

//! Abstract cargo container.
class GCargoContainer : public GTransaction {
 public:
    enum {
        can_Unload     = 1,     ///< this object can provide stuff for unloading
        can_GetUnload  = 2,     ///< this object can receive unloaded stuff
        can_Sup2mc     = 4,     ///< this object can sell supplies

        can_Overload   = 8      ///< do overload. Set by driver code.
    };
 private:
    int32   diffs[el_COUNT];    ///< Target differences.
    int32   reserved[el_COUNT]; ///< Amounts reserved for us. Probably not needed.
    int     flags;
 protected:
    // Interface to underlying object
    /** Return amount this object has (reserved stuff has been subtracted). */
    virtual int32 getAmount(GCargoType type) = 0;
    /** Add some amount to object (may be negative). */
    virtual void changeAmount(GCargoType type, int32 amount) = 0;
    /** Reserve some amount on object (negative = cancel reservation). */
    virtual void reserveAmount(GCargoType type, int32 amount) = 0;
 public:
    GCargoContainer();
    virtual ~GCargoContainer();

    // Public interface to underlying object
    /** Check whether this object can carry the specified cargo type. */
    virtual bool canHave(GCargoType cargo_id) = 0;
    /** Get maximum amount of cargo this object can carry. */
    virtual int32 getMax(GCargoType cargo_id) = 0;
    /** Get name of underlying object. Used for presenting this object
        to the user. */
    virtual string_t getName() const = 0;

    // Public manipulation interface
    int32 getCurrent(GCargoType cargo_id);
    int32 removeCargo(GCargoType type, int32 amount, bool partial);
    int32 addCargo(GCargoType type, int32 amount, bool partial);
    int32 addOrRemoveCargo(GCargoType type, int32 amount, bool partial);

    /** Get container properties. */
    int getFlags() const
        { return flags; }
    /** Set container properties. */
    void setFlags(int fl)
        { flags = fl; }
    /** Check whether container has a certain property. */
    bool hasFlag(int fl) const
        { return flags & fl; }

    // Transaction interface
    void cancel();
    void commit();
};

//! Cargo transfer transaction.
class GCargoTransfer : public GTransaction {
    GCargoContainer* objs[2];
 public:
    GCargoTransfer(GCargoContainer& left, GCargoContainer& right);
    ~GCargoTransfer();

    /** Get object on side.
        \param side which object you want, 0 or 1 */
    GCargoContainer& getContainer(int side)
        { return *objs[side]; }

    int32 move(GCargoType type, int32 amount, int from, bool partial, bool s2m);

    void cancel();
    void commit();
};

//! Transaction which costs some cargo.
class GCargoCostTransaction : public GTransaction {
    enum { LIMIT = el_Torps };

    GCargoContainer& container;

    int32 cost_amounts[LIMIT], reserved_amounts[LIMIT];

    bool isValidIndex(GCargoType t) const;
 protected:
    void changeCost(GCargoType type, int32 amount);
    void setCost(GCargoType type, int32 amount);
    void clearCosts();
 public:
    GCargoCostTransaction(GCargoContainer& container);
    ~GCargoCostTransaction();

    bool isValidCost() const;

    int32 getCost(GCargoType type) const;
    int32 getRemaining(GCargoType type) const;
    int32 getMissing(GCargoType type) const;
    int32 getReserved(GCargoType type) const;

    GError getError();

    void commit();
    void cancel();

    GCargoContainer& getContainer() const
        { return container; }
};

string_t getCargoName(GCargoType type);
string_t getCargoUnit(GCargoType type);

/** Get torpedo type number from cargo type. */
inline int
getTorpTypeFromCargoType(GCargoType type)
{
    ASSERT(type >= el_Torps && type < el_Torps + NUM_TORPS);
    return type - el_Torps + 1;
}

/** Get cargo type from torpedo type.
    \param type torpedo type, [1,NUM_TORPS] */
inline GCargoType
getCargoTypeFromTorpType(int type)
{
    ASSERT(type > 0 && type <= NUM_TORPS);
    return GCargoType(el_Torps + type - 1);
}

#endif
