/**
  *  \file global.h
  *  \brief Global Turn Data
  *
  *  This module provides classes to manage data of a complete turn,
  *  i.e. which contain information about all objects at once.
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GLOBAL_H
#define PCC_V2_GLOBAL_H

#include <map>
#include "io/streams.h"
#include "io/dirs.h"
#include "game/vcrdb.h"
#include "game/msg.h"
#include "game/gen.h"
#include "game/mine.h"
#include "game/cmds.h"
#include "game/gamedirs.h"      // re-exported
#include "cpluslib/sigslot.h"
#include "cpluslib/assert.h"
#include "game/ion.h"
#include "game/player.h"

/**************************** Data Structures ****************************/

class GGameTurn;
class GShip;
class GPlanet;
class GShipType;
class GPlanetType;
class GBaseType;

//! Chart Bounding Box.
struct GChartBBox {
    enum {
        MAX_ZOOM = 10
    };

    GPoint min, max;
    int    max_zoom;
    int    min_zoom;

    void initFromTurn(GGameTurn& trn);
    void initMinZoom();
    void initMaxZoom();
};

//! Flags for GGameTurn::getLocationName().
enum {
    locs_Orbit   = 1,    ///< Show "Orbit of" for planet names.
    locs_Verbose = 2,    ///< Be more verbose.
    locs_Ships   = 4,    ///< Show a ship name if applicable.
    locs_WW      = 8     ///< Show planet name if point is in warp well.
};

//! Everything that represents a game turn.
class GGameTurn {
    /* I'm still unsure what to do with inbox and VCRs */
    struct Player {
        GOutbox           outbox;
        GGen              gen;
        GCommandContainer commands;
        Player(int pid);
        ~Player();
    };
 public:
    typedef std::map<int, GMinefield*> MinefieldMap;
 private:
    GShip*            ships[NUM_SHIPS];
    GPlanet*          planets[NUM_PLANETS];
    GIonStorm*        storms[NUM_STORMS];
    GInbox            inbox;
    Ptr<VcrDatabase>  vcrs;
    MinefieldMap      mines;
    GShipType*        ship_type;
    GPlanetType*      planet_type;
    GBaseType*        base_type;
    GIonStormType*    storm_type;
    GChartBBox        chart_bbox;
    Player*           player_data[NUM_PLAYERS];
    GPlayerSet        playing_set, data_set;
    int               turn_nr;

    void clearAllMines();

    Player& createPlayerData(int player);
    Player& getPlayerDataNC(int player) const
        { return *player_data[player-1]; }
    Player& getPlayerData(int player)
        { return player_data[player-1] ? *player_data[player-1] : createPlayerData(player); }

    void loadPlanetNames();
    void loadStormNames();
public:
    GGameTurn();
    ~GGameTurn();

    void init();
    void done();

    /// \name Player Status Accessors
    //@{
    void setGen(int player, const GGen& gen);
    GGen& getGen(int player)
        { return getPlayerData(player).gen; }
    const GGen& getGen(int player) const
        { return getPlayerDataNC(player).gen; }

    int getTurnNr() const
        { return turn_nr; }
    
    bool isPlaying(int player_nr) const
        { return playing_set.contains(player_nr); }
    bool isData(int player_nr) const
        { return data_set.contains(player_nr); }
    void setHaveData(int player_nr);
    //@}

    /// \name Object Accessors
    //@{
    GShip& getShip(int sid)
        {
            ASSERT(sid > 0 && sid <= NUM_SHIPS);
            return *ships[sid-1];
        }
    const GShip& getShip(int sid) const
        {
            ASSERT(sid > 0 && sid <= NUM_SHIPS);
            return *ships[sid-1];
        }
    GPlanet& getPlanet(int pid)
        {
            ASSERT(pid > 0 && pid <= NUM_PLANETS);
            return *planets[pid-1];
        }
    const GPlanet& getPlanet(int pid) const
        {
            ASSERT(pid > 0 && pid <= NUM_PLANETS);
            return *planets[pid-1];
        }
    GIonStorm& getIonStorm(int sid)
        {
            ASSERT(sid > 0 && sid <= NUM_STORMS);
            return *storms[sid-1];
        }
    const GIonStorm& getIonStorm(int sid) const
        {
            ASSERT(sid > 0 && sid <= NUM_STORMS);
            return *storms[sid-1];
        }
    GMinefield* getMinefield(int id, int cause);
    GMinefield* getMinefield(int id)
        {
            return getMinefield(id, -1);
        }
    MinefieldMap::iterator getMinefieldBegin()
        { return mines.begin(); }
    MinefieldMap::iterator getMinefieldEnd()
        { return mines.end(); }
    //@}

    /// \name Locations
    //@{
    int              getPlanetAt(GPoint pt);
    int              getPlanetAt(GPoint pt, bool gravity);
    int              getShipAt(GPoint pt);
    int              getGravityPlanetAt(GPoint pt);
    int              getNumberOfPlayedShipsAt(GPoint pt);
    string_t         getLocationName(GPoint pt, int flags);

    GChartBBox& getChartBBox()
        { return chart_bbox; }
    //@}

    /// \name Message Accessors
    //@{
    GInbox& getInbox()
        { return inbox; }
    GInbox const& getInbox() const
        { return inbox; }
    GOutbox const& getOutbox(int player) const
        { return getPlayerDataNC(player).outbox; }
    GOutbox& getOutbox(int player)
        { return getPlayerData(player).outbox; }
    Ptr<VcrDatabase> getVcrs() const
        { return vcrs; }
    GCommandContainer& getCommands(int player)
        { return getPlayerDataNC(player).commands; }
    //@}

    /// \name Meta-Information
    //@{
    GShipType& getShipType()
        { return *ship_type; }
    GPlanetType& getPlanetType()
        { return *planet_type; }
    GBaseType& getBaseType()
        { return *base_type; }
    GIonStormType& getIonStormType()
        { return *storm_type; }
    //@}

    /// \name Loading
    //@{
    void loadPlanetXY(Stream&);
    void postprocess();
    //@}

    /// \name Status Updates
    //@{
    void doUpdates();

    /// Signal: Planet was changed. Parameter is planet Id.
    Signal1<int> sig_planet_changed;
    /// Signal: Ship was changed. Parameter is ship Id.
    Signal1<int> sig_ship_changed;
    //@}
};

/******************************* Accessors *******************************/

// These are exposed here only for the following inline functions
extern Ptr<GGameTurn> current_turn_internal;
extern Ptr<GGameTurn> display_turn_internal;

/** Get global turn instance. */
inline GGameTurn&  getCurrentTurn()   { return *current_turn_internal; }
inline GGameTurn&  getDisplayedTurn() { return *display_turn_internal; }

void setDisplayedTurn(Ptr<GGameTurn> disp);

extern Signal0 sig_turn_changed;

/** Global Data. */
void initTurn();
void initGameGlobals(string_t gamedir, string_t rootdir);

#endif
