/**
  *  \file game/mission.cc
  *  \brief Starship Mission Management
  *
  *  This module defines functions to manage ship mission definitions.
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include <iostream>
#include <cctype>
#include "cpluslib/format.h"
#include "game/mission.h"
#include "game/player.h"
#include "game/ship.h"
#include "util/console.h"
#include "util/except.h"
#include "util/misc.h"
#include "util/nls.h"
#include "io/streams.h"
#include "io/iostr.h"
#include "game/gamedirs.h"
#include "game/config.h"
#include "game/host.h"

/** Construct blank mission definition. */
GMission::GMission()
    : number(0), flags(0), hotkey(0)
{
    arg_spec[0] = arg_spec[1] = 0;
}

/** Construct mission definition from mission.cc line. */
GMission::GMission(int number, string_t desc_line)
    : number(number), race_mask(GPlayerSet::allUpTo(NUM_OWNERS)), flags(0), hotkey(0)
{
    arg_spec[0] = arg_spec[1] = 0;
    parseDescription(desc_line);
}

GMission::~GMission()
{ }

/** True iff mission (might) work on the specified ship. Actually,
    this should return true if we want the user to be able to set the
    mission on the specified ship. */
bool
GMission::worksOn(GShip& ship) const
{
    if (! race_mask.contains(raceId(ship.getRealOwner())))
        return false;
    if ((flags & mf_Registered) && !allowRegistered(ship.getRealOwner()))
        return false;

    // FIXME: check more conditions

    return true;
}

/** Get name for this mission if it is set on /ship/. This name is
    intended to be displayed on the ship screen.

    NOTE: This function intentionally does not take a 'const GShip',
    to reserve the possibility to call the interpreter from here
    (which might need a non-const ship). */
string_t
GMission::getName(GShip& ship) const
{
    string_t s = getName();
    if (ship.getIntId())
        s += format(" I:%d", ship.getIntId());
    if (ship.getTowId())
        s += format(" T:%d", ship.getTowId());
    return s;
}

/** Get name for parameter. */
string_t
GMission::getArgName(Parameter id) const
{
    if (arg_names[id].empty()) {
        if (id)
            return "Tow";
        else
            return "Intercept";
    }
    return arg_names[id];
}

/** Parse mission.cc description line. This will set all affected
    members according to that definition.

    \param desc_line line from mission.cc, sans mission number
    (i.e. "!h#,Tow a ship") */
void
GMission::parseDescription(const string_t desc_line)
{
    string_t::size_type n = 0;

    int arg_type = mat_Integer, arg_flags = 0;
    while (n < desc_line.length() && desc_line[n] != ',') {
        switch (desc_line[n++]) {
         case '+':
         case '-':
            {
                bool negate = (desc_line[--n] == '-');
                if (!negate)
                    race_mask = GPlayerSet();
                while (++n < desc_line.length() && desc_line[n] != ',') {
                    int p = getPlayerIdFromChar(desc_line[n]);
                    if (p > 0) {
                        if (negate)
                            race_mask -= p;
                        else
                            race_mask += p;
                    }
                }
            }
            break;
         case 'r': flags |= mf_Registered; break;
         case 'i': flags |= mf_Waypoint; break;
         case 'o': arg_flags |= ma_Own; break;
         case '!': arg_flags |= ma_NotThis; break;
         case 'p': arg_type = mat_PID; break;
         case 's': arg_type = mat_SID; break;
         case 'h': arg_type = mat_SID_here; break;
         case 'b': arg_type = mat_BID; break;
         case 'y': arg_type = mat_Player; break;
         case '*':
            arg_spec[0] = arg_type | arg_flags;
            break;
         case '#':
            arg_spec[1] = arg_type | arg_flags;
            break;
        }
    }
    if (n < desc_line.length()) {
        name = desc_line.substr(n+1);

        /* assign hot-key if so desired */
        string_t::size_type p = name.find('~');
        if (p != name.npos && p+1 < name.length()) {
            using namespace std;
            hotkey = tolower((unsigned char) name[p+1]);
            name.erase(p, 1);
        }
    }
}

/****************************** GMissionList *****************************/

/// Global instance of the mission table.
GMissionList GMissionList::instance;

/** Construct empty mission list. */
GMissionList::GMissionList()
    : used_letters(~0U)
{ }

/** Construct mission list for a ship. The list will contain all missions
    doable by the specified ship.
    \param other complete mission definition
    \param ship  ship for which to prepare the list */
GMissionList::GMissionList(const GMissionList& other, GShip& ship)
    : used_letters(~0U)
{
    for (iterator i = other.begin(); i != other.end(); ++i)
        if (i->worksOn(ship))
            data.push_back(*i);
}

/** Destructor. */
GMissionList::~GMissionList()
{ }

/** Add mission to list. This will add a (copy of) /msn/ to the
    mission list. If the mission was already defined, the call is
    ignored.
    \return true iff mission was added, false if call was ignored */
bool
GMissionList::addMission(const GMission& msn)
{
    /* do not add if it's already there. A mission is considered
       already there if its race mask is a superset of ours. */
    for (iterator i = begin(); i != end(); ++i) {
        if (i->getNumber() != msn.getNumber())
            continue;
        if ((msn.getRaceMask() - i->getRaceMask()).nonempty())
            continue;
        return false;
    }

    /* assign hot-key. Do this without modifying /msn/. */
    char c = msn.getHotkey();
    if (c == 0) {
        if (msn.getNumber() >= 0 && msn.getNumber() < 10) {
            c = '0' + msn.getNumber();
        } else {
            for (int i = 0; i < 26; ++i) {
                if (!(used_letters & (1 << i))) {
                    c = 'a' + i;  // FIXME: assumes ASCII
                    break;
                }
            }
            if (c == 0) {
                /* all letters used up */
                used_letters = 0;
                c = 'a';
            }
        }
    }
    ASSERT(c);
    if (c >= 'a' && c <= 'z')
        used_letters |= 1 << (c - 'a');

    data.push_back(msn);
    data.back().setHotkey(c);
    return true;
}

static bool
missionCompare(const GMission& a, const GMission& b)
{
    if (a.getNumber() < b.getNumber())
        return true;
    if (a.getNumber() > b.getNumber())
        return false;
    return a.getRaceMask().toInteger() < b.getRaceMask().toInteger();
}

/** Sort mission list. Brings list into numerical mission order, which
    will then be used on the mission screen. */
void
GMissionList::sort()
{
    std::sort(data.begin(), data.end(), missionCompare);
}

/** Clear this mission list instance. */
void
GMissionList::clear()
{
    data.clear();
}

/** Get mission by number/player.
    \param id mission number
    \param race_mask races (not players!) for which you want the mission. */
const GMission*
GMissionList::getMissionByNumber(int id, GPlayerSet race_mask) const
{
    for (iterator i = begin(); i != end(); ++i)
        if (i->getNumber() == id && (i->getRaceMask() & race_mask).nonempty())
            return &*i;
    return 0;
}

/** Load mission.cc file.
    \param in      stream
    \param name    name for that stream, for users. */
void
GMissionList::loadFromFile(std::istream& in, const string_t& name)
{
    int line_nr = 0;
    string_t line;
    bool have_mission = false;

    while (getline(in, line)) {
        ++line_nr;
        line = strTrim(line);
        if (line.empty() || line[0] == ';')
            continue;

        string_t::size_type p = line.find_first_of("=,");
        if (p == string_t::npos)
            console.write(LOG_ERROR, format(_("%s:%d: missing delimiter"), name, line_nr));
        if (line[p] == ',') {
            int mnum;
            if (!stringValue(strTrim(line.substr(0, p)), mnum) || mnum < 0 || mnum > 10000) {
                console.write(LOG_ERROR, format(_("%s:%d: invalid mission number")));
                continue;       // with next mission
            }

            line.erase(0, p+1);
            try {
                have_mission = addMission(GMission(mnum, line));
            }
            catch(UserException& e) {
                console.write(LOG_ERROR, format("%s:%d: %s", name, line_nr, e.what()));
            }
        } else {
            /* additional assignments. */
            if (!have_mission)
                continue;

            string_t lhs = strTrim(line.substr(0, p));
            line = strTrim(line.substr(p+1));
            if (stringMatch("I", lhs)) {
                data.back().setArgName(GMission::InterceptParameter, line);
            } else if (stringMatch("J", lhs)) {
                data.back().setArgName(GMission::TowParameter, line);
            } else if (stringMatch("Shortname", lhs)) {
                // short mission name
            } else if (stringMatch("Condition", lhs)) {
                // condition which says whether this mission is allowed
                // to be set on this ship
            } else if (stringMatch("Text", lhs)) {
                // Text to show if mission is set
            } else if (stringMatch("Willwork", lhs)) {
                // condition which says whether the mission will work,
                // to warn the user when he sets cloaking on a damaged
                // ship.
            }
        }
    }
}

/** Initialize missions. This loads the mission definition files into
    the global GMissionList instance. It is idempotent, i.e. can be
    called at any time. */
void
initMissions()
{
    GMissionList::getInstance().clear();
    Ptr<Stream> s = spec_file_dir->openFileNT("missions.usr", Stream::C_READ);
    if (s) {
        StandardStream inp(s);
        GMissionList::getInstance().loadFromFile(inp, s->getName());
    }
    s = spec_file_dir->openFileNT("mission.cc", Stream::C_READ);
    if (s) {
        StandardStream inp(s);
        GMissionList::getInstance().loadFromFile(inp, s->getName());
    }

    GMissionList::getInstance().sort();
}

/** Check whether a mission causes the ship to cloak. */
bool
isMissionCloaking(int mission_id, int owner)
{
    // you thought Lisp has lots-a-parens?
    return (mission_id == GMission::msn_Cloak
            || (host.isPHost() && mission_id == config.ExtMissionsStartAt + GMission::pmsn_Cloak)
            || (raceMissionId(owner) == 3
                && (mission_id == GMission::msn_Special
                    || (host.isPHost()
                        && (mission_id == config.ExtMissionsStartAt + GMission::pmsn_Special
                            || mission_id == config.ExtMissionsStartAt + GMission::pmsn_StandardSuperSpy)))));
}
