/**
  *  \file game/planet-form.cc
  *  \brief Planet-related Formulas
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include "game/planet.h"
#include "util/nls.h"
#include "cpluslib/misc.h"
#include "game/config.h"
#include "game/formula.h"
#include "game/host.h"

/** Get name for a temperature.
    \param nr government number
    \return statically-allocated, translated string. */
const char*
GPlanet::getTempStr(int temp)
{
    if (temp <= 14)
        return _("arctic");
    else if (temp <= 39)
        return _("cool");
    else if (temp <= 64)
        return _("warm");
    else if (temp <= 84)
        return _("tropical");
    else
        return _("desert");
}

/** Get name for a native race.
    \param nr government number
    \return statically-allocated, translated string. */
const char*
GPlanet::getNativeRaceName(int nr)
{
    static const char*const names[] = {
        N_("Humanoid"),
        N_("Bovinoid"),
        N_("Reptilian"),
        N_("Avian"),
        N_("Amorphous"),
        N_("Insectoid"),
        N_("Amphibian"),
        N_("Ghipsoldal"),
        N_("Siliconoid")
    };
    if (nr <= 0 || nr > int(countof(names)))
        return "???";
    else
        return _(names[nr-1]);
}

/** Get name for a native government.
    \param nr government number
    \return statically-allocated, translated string. */
const char*
GPlanet::getNativeGovernmentName(int nr)
{
    static const char*const names[] = {
        N_("Anarchy"),
        N_("Pre-Tribal"),
        N_("Early-Tribal"),
        N_("Tribal"),
        N_("Feudal"),
        N_("Monarchy"),
        N_("Representative"),
        N_("Participatory"),
        N_("Unity")
    };
    if (nr <= 0 || nr > int(countof(names)))
        return "???";
    else
        return _(names[nr-1]);
}

/** Get population happiness string.
    \param happy happiness value
    \return statically-allocated, translated string. */
const char*
GPlanet::getHappyStr(int happy)
{
    if (happy >= 90)
        return _("happy");
    else if (happy >= 70)
        return _("calm");
    else if (happy >= 50)
        return _("unhappy");
    else if (happy >= 40)
        return _("very angry");
    else if (happy >= 20)
        return _("rioting");
    else
        return _("fighting");
}

/** Get population happiness change string.
    \param change happiness change
    \return statically-allocated, translated string. */
const char*
GPlanet::getChangeStr(int change)
{
    if (change < -5)
        return _("They HATE you!");
    else if (change < 0)
        return _("They are angry about you!");
    else if (change == 0)
        return _("They are undecided about you.");
    else if (change > 4)
        return _("They LOVE you.");
    else
        return _("They like your leadership.");
}

/** Get starbase mission name.
    \param mission mission number
    \return statically-allocated, translated string. */
const char*
GPlanet::getBaseMissionName(int mission)
{
    switch (mission) {
     case 0:
        return _("none");
     case 1:
        return _("Refuel ships");
     case 2:
        return _("Maximize defense");
     case 3:
        return _("Load torpedoes onto ships");
     case 4:
        return _("Unload incoming ships");
     case 5:
        return _("Repair base");
     case 6:
        return _("Force surrender");
     default:
        return "???";
    }
}

static inline double
pdiv(int a, int b)
{
    if (host.isPHost())
        return double(a) / b;
    else
        return a / b;
}

/** Native happiness change.
    \param tax   desired tax rate
    \param mifa  mines plus factories on planet */
int
GPlanet::getNativeChange(int tax, int mifa) const
{
    /* formula is correct for PHost and THost */
    int change =
        int32(500 + 50 * getNativeGovernment()
              - 85 * tax
              - pdiv(mifa, 2)
              - std::sqrt(double(getNatives())))
        / 100;
    if (getNativeRace() == nr_Avian)
        change += 10;
    return change;
}

/** Native happiness change.
    \returns happiness change for current situation. */
int
GPlanet::getNativeChange() const
{
    return getNativeChange(getNativeTax(), getStructures(str_Mines) + getStructures(str_Factories));
}

/** Colonists change.
    \param tax   desired tax rate
    \param mifa  mines plus factories on planet */
int
GPlanet::getColonistChange(int tax, int mifa) const
{
    double common = 1000 - 80*tax - std::sqrt(double(getColonists()));
    bool   crystal = (raceId(getOwner()) == 7) && config.CrystalsPreferDeserts;
    if (host.isPHost()) {
        if (crystal)
            return int32(common - mifa/3.0 - (100-getTemp())/0.66) / 100;
        else
            return int32(common - mifa/3.0 - std::abs(getTemp()-50)/0.33) / 100;
    } else {
        if (crystal)
            return int32(common - (mifa/3 + 3*(100-getTemp()))) / 100;
        else
            return int32(common - (mifa/3 + 3*std::abs(getTemp()-50))) / 100;
    }
}

/** Colonists happiness change on this planet.
    \returns change on this planet in current situation */
int
GPlanet::getColonistChange() const
{
    return getColonistChange(getColonistTax(), getStructures(str_Mines) + getStructures(str_Factories));
}

/** Maximum number of structures.
    \param clans      clans on planet
    \param threshold  below threshold, growth is linear; above it's rootic :-) */
int
GPlanet::getMaxStructuresFormula(int32 clans, int threshold)
{
    if (clans <= threshold)
        return clans;
    else
        /* the rounding function is actually ERnd, but that never
           makes a difference; sqrt(INT) never ends in .5 exactly */
        return threshold + int(std::sqrt(double(clans - threshold)) + 0.5);
}

int
GPlanet::getMaxStructures(TPlanetaryStructure ps, int32 clans, bool base)
{
    switch (ps) {
     case str_Factories:
        return getMaxStructuresFormula(clans, 100);
     case str_Mines:
        return getMaxStructuresFormula(clans, 200);
     case str_Defense:
        return getMaxStructuresFormula(clans, 50);
     case str_BaseDefense:
        return base ? 200 : 0;
    }
    return 0;
}

/** Maximum population on planet, for a race.
    \param race player number
    \returns maximum population in clans. */
int32
GPlanet::getMaxSupportedColonists(int player) const
{
    bool crystal = (raceId(player) == 7 && config.CrystalsPreferDeserts);
    int32 limit;
    if (host.isPHost()) {
        if (!config.ClimateLimitsPopulation)
            return 250000;
        if (crystal) {
            if (config.CrystalSinTempBehavior)
                if (getTemp() >= 15)
                    return int32(100000 * std::sin(getTemp() * PI / 100));
                else
                    return 3 + getTemp() * config.MaxColTempSlope / 100;
            else
                return 1000 * getTemp();
        }

        if (getTemp() >= 85)
            limit = config.MaxColTempSlope * (100-getTemp()) / 100 + 1;
        else if (getTemp() <= 14)
            limit = config.MaxColTempSlope * getTemp() / 100 + 3;
        else
            limit = int32(100000 * std::sin(getTemp() * PI / 100));
    } else {
        if (!config.ClimateLimitsPopulation)
            return 100000;
        if (crystal)
            return 1000 * getTemp();

        if (getTemp() >= 85)
            limit = 2 * (100 - getTemp()) + 1;
        else if (getTemp() <= 14)
            limit = 2 * getTemp() + 3;
        else
            limit = int32(100000 * std::sin((100 - getTemp()) * 0.0314) + 0.5);
    }

    if (raceId(player) == 10 && getTemp() <= 19 && limit < 90000)
        limit = 90000;
    if ((raceId(player) == 4 || raceId(player) >= 9) && getTemp() >= 84 && limit < 60)
        limit = 60;

    return limit;
}

/** Get colonist tax amount, as requested. This returns the amount we want
    from the colonists, not the amount we'll finally get. */
int32
GPlanet::getColonistDue(const int tax) const
{
    // Note that these formulas differ in rounding only. PHost uses
    // `Round', THost uses `ERnd' aka `I-don't-care-how-it-rounds'.
    const int rate = config.ColonistTaxRate[playerIndex(getOwner())];
    if (host.isPHost()) {
        return radiv(radiv(getColonists() * tax, 1000) * rate, 100);
    } else {
        return rdivaddl(rdivaddl(getColonists() * tax, 1000, 0) * rate, 100, 0);
    }
}

/** Get colonist tax income.
    \param tax     [in] tax rate
    \param rem_inc [out] this is set to the amount of taxes which we can
    still get from the natives (regarding MaxPlanetaryIncome)
    \returns megacredits collected. */
int32
GPlanet::getColonistDueLimited(const int tax, int32& rem_inc) const
{
    const int32 due = getColonistDue(tax), max = config.MaxPlanetaryIncome;
    if (due < max) {
        rem_inc = max - due;
        return due;
    } else {
        rem_inc = 0;
        return max;
    }
}

/** Get native tax amount, as requested, for this planet. This returns
    the amount we're asking from them, not what we'll get. */
int32
GPlanet::getNativeDue(int tax) const
{
    return getNativeDue(tax, getNativeRace(), getNativeGovernment(),
                        getNatives(), getOwner());
}

/** Get native tax amount, as requested, parameterized. This returns
    the amount we're asking from them, not what we'll get.
    \param tax    tax rate
    \param race   native race
    \param gov    government factor (SPI)
    \param pop    native clans
    \param owner  owner (for production rates) */
int32
GPlanet::getNativeDue(int tax, int race, int gov, int32 pop, int owner)
{
    int32 due;
    if (host.isPHost()) {
        due = radiv(radiv(tax * gov * pop, 5000) * config.NativeTaxRate[playerIndex(owner)], 100);
    } else {
        due = rdivaddl(rdivaddl(tax * gov * pop, 5000, 0) * config.ColonistTaxRate[playerIndex(owner)], 100, 0);
    }
    if (race == nr_Insectoid)
        return 2*due;
    else
        return due;
}

/** Get native tax amount, limited. Limits the tax income by income
    limit and available colonists.
    \param tax     tax rate
    \param rem_inc remaining allowed income; output of getColonistDueLimited(). */
int32
GPlanet::getNativeDueLimited(int tax, int32 rem_inc) const
{
    /* amorphs don't pay */
    if (getNativeRace() == nr_Amorphous)
        return 0;

    /* cyborgs can only tax to 20 */
    int limit = host.getNativeTaxRateLimit(getOwner());
    if (tax > limit)
        tax = limit;

    /* normal formulas here */
    int32 due = getNativeDue(tax);
    int32 colos = getColonists();

    if (host.isPHost()) {
        if (getNativeRace() == nr_Insectoid)
            colos *= 2;
        colos = radiv(colos * config.NativeTaxRate[playerIndex(getOwner())], 100);
    } else {
        colos = colos * config.ColonistTaxRate[playerIndex(getOwner())] / 100;
        if (getNativeRace() == nr_Insectoid)
            colos *= 2;
    }

    if (due > colos)
        due = colos;
    if (due > rem_inc)
        due = rem_inc;
    return due;
}

/** Get Amorphous colonist breakfast. Returns the number of clans
    eaten by our lovely amorphous natives when they have a happiness
    as specified. */
int32
GPlanet::getAmorphousBreakfast(int happy)
{
    if (host.isPHost()) {
        if (happy >= 70)
            return 5;
        else if (happy >= 50)
            return 20;
        else
            return 40;
    } else {
        if (happy > 90)
            return 5;
        else
            return 95 - happy;
    }
}

/** Get Bovinoid supply contribution. Contribution is not yet limited by
    colonists. */
int32
GPlanet::getBovinoidSupplyContribution() const
{
    if (getNativeRace() == nr_Bovinoid)
        return getBovinoidSupplyContribution(getNatives(), getOwner());
    else
        return 0;
}

/** Get Bovinoid supply contribution. This returns the amount of
    supplies finally collected. */
int32
GPlanet::getBovinoidSupplyContributionLimited() const
{
    // FIXME: the same problem as in getBovinoidSupplyContribution
    // appears here. We *could* add a workaround here (compute totals,
    // compute factory contribution, compute difference), but is it
    // really worth to do that for an accuracy of +/-1 with a config
    // option no-one uses?
    //      -- update: it's going to be used in PList 2.5, so someday
    //         it should be supported
    if (getNativeRace() == nr_Bovinoid) {
        int32 limit = host.isPHost() ? getColonists() * config.ProductionRate[playerIndex(getOwner())] / 100 : getColonists();
        int32 due = getBovinoidSupplyContribution();
        if (due < limit)
            return due;
        else
            return limit;
    } else {
        return 0;
    }
}

/** Get Bovinoid supply contribution, parameterized. */
int32
GPlanet::getBovinoidSupplyContribution(int32 pop, int owner)
{
    // FIXME: for ultimate accuracy, we have to consider factories as
    // well. PHost does "(bovi + factories) * ProductionRate", not
    // "bovi * ProductionRate + factories * ProductionRate".
    if (host.isPHost())
        return (pop / 100) * config.ProductionRate[playerIndex(owner)] / 100;
    else
        return pop / 100;
}
