/**
  *  \file game/save.cc
  *  \brief Saving Game Data -- General Utilities
  *
  *  This module contains the basic routines for writing out object
  *  lists.
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/save.h"
#include "game/global.h"
#include "game/ship.h"
#include "io/io.h"
#include "game/checksum.h"
#include "game/planet.h"

/** Save ships to stream. Writes a "count" word plus ship records.
    \param trn       turn to save
    \param str       stream to save in
    \param src       source flag mask to use
    \param checksum  pointer to checksum word to update, or null
    \param flags     flags (save_XXX) */
void
saveShips(GGameTurn& trn, Stream& str, GPlayerSet src, uint32* checksum, int flags)
{
    /* pass 1: count */
    TInt16 v = { 0 };
    for (int i = 1; i <= NUM_SHIPS; ++i)
        if (trn.getShip(i).isExistant() && trn.getShip(i).getSrcFlags().containsAnyOf(src))
            ++v.value;
    storeStructure(str, v);
    if ((flags & save_ChecksumCount) && checksum)
        *checksum += getChecksum(v.value);

    /* pass 2: data */
    char buffer[TShip::size];
    for (int i = 1; i <= NUM_SHIPS; ++i) {
        if (trn.getShip(i).isExistant() && trn.getShip(i).getSrcFlags().containsAnyOf(src)) {
            TShip tmp;
            trn.getShip(i).getRawData(tmp);
            storeStructure(buffer, tmp);
            str.write(buffer, sizeof buffer);

            uint32 csum = getChecksum(buffer, sizeof buffer);
            setControlEntry(1, i, csum);
            if (checksum)
                *checksum += csum;
        }
    }
}

/** Save planets to stream. Writes a "count" word plus planet records.
    \param trn       turn to save
    \param str       stream to save in
    \param src       source flag mask to use (generally 1<<pid)
    \param checksum  pointer to checksum word to update, or null
    \param flags     flags (save_XXX) */
void
savePlanets(GGameTurn& trn, Stream& str, GPlayerSet src, uint32* checksum, int flags)
{
    /* pass 1: count */
    TInt16 v = { 0 };
    for (int i = 1; i <= NUM_PLANETS; ++i)
        if (trn.getPlanet(i).isExistant() && trn.getPlanet(i).getPlanetSrcFlags().containsAnyOf(src))
            ++v.value;
    storeStructure(str, v);
    if ((flags & save_ChecksumCount) && checksum)
        *checksum += getChecksum(v.value);

    /* pass 2: data */
    char buffer[TPlanet::size];
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        if (trn.getPlanet(i).isExistant() && trn.getPlanet(i).getPlanetSrcFlags().containsAnyOf(src)) {
            TPlanet tmp;
            trn.getPlanet(i).getRawData(tmp);
            storeStructure(buffer, tmp);
            str.write(buffer, sizeof buffer);

            uint32 csum = getChecksum(buffer, sizeof buffer);
            setControlEntry(2, i, csum);
            if (checksum)
                *checksum += csum;
        }
    }
}

/** Save bases to stream. Writes a "count" word plus base records.
    \param trn       turn to save
    \param str       stream to save in
    \param src       source flag mask to use (generally 1<<pid)
    \param checksum  pointer to checksum word to update, or null
    \param flags     flags (save_XXX) */
void
saveBases(GGameTurn& trn, Stream& str, GPlayerSet src, uint32* checksum, int flags)
{
    /* pass 1: count */
    TInt16 v = { 0 };
    for (int i = 1; i <= NUM_PLANETS; ++i)
        if (trn.getPlanet(i).isExistant()
            && trn.getPlanet(i).hasBase()
            && trn.getPlanet(i).getBaseSrcFlags().containsAnyOf(src))
            ++v.value;
    storeStructure(str, v);
    if ((flags & save_ChecksumCount) && checksum)
        *checksum += getChecksum(v.value);

    /* pass 2: data */
    char buffer[TStarbase::size];
    for (int i = 1; i <= NUM_PLANETS; ++i) {
        if (trn.getPlanet(i).isExistant()
            && trn.getPlanet(i).hasBase()
            && trn.getPlanet(i).getBaseSrcFlags().containsAnyOf(src))
        {
            TStarbase tmp;
            trn.getPlanet(i).getRawBaseData(tmp);
            storeStructure(buffer, tmp);
            str.write(buffer, sizeof buffer);

            uint32 csum = getChecksum(buffer, sizeof buffer);
            setControlEntry(3, i, csum);
            if (checksum)
                *checksum += csum;
        }
    }
}
