/**
  *  \file game/vcrdb.h
  *  \brief VCR Database
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_VCRDB_H
#define PCC_V2_GAME_VCRDB_H

#include <vector>
#include "game/vcr.h"
#include "io/streams.h"

/// High-level representation of a VCR
class VcrEntry {
    int   status;               ///< Battle outcome (VCRS_xxx).
    int   hulls[2];             ///< Hulls of each side, if known.
    TVcr  vcr;                  ///< Starting parameters.
    TVcr  result;               ///< Result, if known.

    int   guessHull(int side) const;
 public:
    VcrEntry(const TVcr& vcr);

    void setResultFromPlayer(VcrPlayer& player);
    void setResult(int status, const TVcr& result);

    /** Get combat recording in "raw" format. */
    const TVcr& getVcr() const
        { return vcr; }
    /** Get result of combat in "raw" format. */
    const TVcr& getResult() const
        { return result; }
    /** Get status of fight, if known. */
    int getStatus() const
        { return status; }
    /** Get guessed hull, if known.
        \return hull number, zero if not known. */
    const int getGuessedHull(int side) const
        { return hulls[side]; }

    bool canBeHull(int side, int hull_id) const;

    /** Check whether \c side is a planet. */
    bool isPlanet(int side) const
        { return vcr.battle_type && side; }
};

/// VCR Database
class VcrDatabase {
    // FIXME: I don't particularly like that the player is "hardwired"
    // in the database. In particular, I'd like to get rid of the
    // Ptr<>.

    typedef std::vector<VcrEntry*> entry_vec_type;

    entry_vec_type contents;
    Ptr<VcrPlayer> vcr_player;
    uint16 capabilities;

    enum TVcrType {
        vcr_None,               ///< Unknown.
        vcr_Host,               ///< Tim's Host.
        vcr_PHost2,             ///< PHost 2.
        vcr_PHost3,             ///< PHost 3, or PHost 4 without new features.
        vcr_PHost4              ///< PHost 4.
    };
    TVcrType vcr_type;
 public:
    VcrDatabase();
    ~VcrDatabase();

    void loadFromFile(Stream& file);

    void addVcr(const TVcr& avcr);

    /** Get Number of VCRs in Database. */
    int getVcrCount() const
        { return contents.size(); }

    /** Return VCR database entry, 0-based. */
    VcrEntry& getEntry(int index)
        { return *contents[index]; }

    /** Get capabilities flag word from VCR data file. */
    unsigned getCapabilities() const
        { return capabilities; }

    void setVcrPlayer(Ptr<VcrPlayer> avp);

    Ptr<VcrPlayer> getVcrPlayer();
};

#endif
