/**
  *  \file gfx/sprite3d.cc
  *  \brief Sprite Management
  *
  *  This module provides classes to manage sprites in a semi-3D
  *  environment. Unlike the old GfxSprite/GfxSpriteManager pair, this
  *  allows for arbitrary Z values and sprites which change their Z
  *  coordinate during play.
  *
  *  (c) 2001-2005 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include <iostream>
#include "gfx/sprite3d.h"
#include "gfx/multiclip.h"
#include "gfx/palette.h"
#include "gfx/clip.h"

/****************************** GfxSprite3D ******************************/

/*! \class GfxSprite3D
    This represents a sprite in pseudo-3-D space. In addition to X/Y
    coordinates, which map 1:1 to screen coordinates, there's a Z
    coordinate which describes the layering of the sprite. Z
    coordinate zero is on top and always visible, higher values are
    farther behind (and possibly obscured), lower values are not
    visible.

    Each sprite has a user-defined identifier. This identifier can be
    used to find sprites according to various criteria. In addition,
    identifiers decide sort order in case of equal Z coordinates: low
    Ids sort before high Ids. There can be many sprites with the same
    Id. */

/** Every-tick action. This function is called by GfxSpriteManager3D,
    do not call it yourself.
    \default does nothing */
void
GfxSprite3D::tick()
{ }

/** Delete this sprite. This only marks the sprite for deletion, it
    will deleted upon the next call to GfxSpriteManager3D::doUpdates(). */
void
GfxSprite3D::deleteThis()
{
    if (!md.need_delete && md.manager) {
        md.need_delete = true;
        md.manager->need_delete = true;
    }
}

/** Update this sprite. This tells GfxSpriteManager3D::doUpdates() that
    this sprite wants to be redrawn. */
void
GfxSprite3D::update()
{
    if (!md.need_update && md.manager) {
        md.need_update = true;
        md.manager->need_update = true;
    }
}

/** Set position of this sprite. Causes all necessary redisplay
    operations to be scheduled.
    \param area new screen position
    \param z new Z coordinate */
void
GfxSprite3D::setPos(GfxRect area, int z)
{
    if (!md.manager) {
        this->pos_2d = area;
        this->pos_z  = z;
        return;
    }

    if (area != this->pos_2d) {
        if (!md.was_moved) {
            md.old_area = this->pos_2d;
            md.was_moved = true;
        }
        this->pos_2d = area;
        update();
    }
    if (z != this->pos_z) {
        md.manager->need_sort = true;
        this->pos_z = z;
        update();
    }
}

/** Set identifier.
    \param id new identifier */
void
GfxSprite3D::setId(int id)
{
    if (id != this->identifier) {
        this->identifier = id;
        if (md.manager)
            md.manager->need_sort = true;
    }
}

/** Set position in 2-D space. This modifies the position of the
    sprite's left/top corner, without affecting its size or Z
    position.
    \param pt new top-left position
    \see setPos */
void
GfxSprite3D::setPos2D(GfxPoint pt)
{
    setPos(GfxRect(pt.x, pt.y, pos_2d.w, pos_2d.h), pos_z);
}

/** Set position in 2-D space. This modifies the position of the
    sprite's center point, without affecting its size or Z position.
    \param pt new center position
    \see setPos */
void
GfxSprite3D::setCenterPos2D(GfxPoint pt)
{
    setPos(GfxRect(pt.x - pos_2d.w/2, pt.y - pos_2d.h/2, pos_2d.w, pos_2d.h), pos_z);
}

/** Set center position in 3-D space. This modifies the position of the
    sprite's center point.
    \param pt new center position
    \param z new Z position
    \see setPos */
void
GfxSprite3D::setCenterPos(GfxPoint pt, int z)
{
    setPos(GfxRect(pt.x - pos_2d.w/2, pt.y - pos_2d.h/2, pos_2d.w, pos_2d.h), z);
}

/** Set associated sprite manager. Called by the sprite manager when
    the sprite is added/removed.
    \param manager new sprite manager */
void
GfxSprite3D::setManager(GfxSpriteManager3D* manager)
{
    md.manager = manager;
}

/** Ensure this sprite was ticked. Calls tick() if it was not already
    called in this GfxSpriteManager3D::tick() run. Calling this
    outside a GfxSpriteManager3D::tick() has no effect. */
void
GfxSprite3D::ensureTick()
{
    if (md.manager && md.tick_mark != md.manager->tick_mark) {
        md.tick_mark = md.manager->tick_mark;
        tick();
    }
}

/*************************** GfxSpriteManager3D **************************/

/*! \class GfxSpriteManager3D

    This class manages a set of sprites. It provides useful accessors
    and functions to update display silently. */

/** Constructor.
    \param bbox bounding box. All output will be clipped at this box. */
GfxSpriteManager3D::GfxSpriteManager3D(GfxRect bbox)
    : sprites(), bbox(bbox), need_sort(false), need_delete(false), need_update(false),
      tick_mark(false)
{ }

/** Destructor. */
GfxSpriteManager3D::~GfxSpriteManager3D()
{ }

static bool
sortByDepth(GfxSprite3D* a, GfxSprite3D* b)
{
    if (a->getPosZ() == b->getPosZ())
        return a->getId() > b->getId();
    else
        return a->getPosZ() > b->getPosZ();
}

/** Call tick() on all sprites. To make the effects actually
    displayed, use
    - doUpdates() for a normal (optimized) update
    - doUpdatesSilently() followed by draw() for a brute-force redraw.
      If you called tick() many times and expect many changes, this is
      expected to be faster than normal doUpdates(). */
void
GfxSpriteManager3D::tick()
{
    tick_mark = !tick_mark;

    /* Note that we cannot use iterators here since
       tick() might call addSprite() which invalidates iterators.  */
    for (container_t::size_type i = 0; i < sprites.size(); ++i)
        if (! sprites[i]->md.need_delete)
            sprites[i]->ensureTick();
}

static GfxRect
expandRectangle(GfxRect r)
{
#if 1
    /* round up all rectangles to a multiple of 64. This speeds up
       display quite a lot (FLAK with >300 ships goes from 1 fps to 20
       fps or so) */
    GfxRect rv;
    static const int scale = 64;
    static const int mask = ~(scale-1);

    rv.x = r.x & mask;
    rv.y = r.y & mask;
    rv.w = ((r.x + r.w + scale-1) & mask) - rv.x;
    rv.h = ((r.y + r.h + scale-1) & mask) - rv.y;

    return rv;
#else
    return r;
#endif
}

/** Handle updates. Assuming /can/ still displays what the last
    doUpdates() or draw() left over, displays the current situation
    \param can canvas to operate on. doUpdates() will perform
    locking and update
    \see doUpdatesSilently() */
void
GfxSpriteManager3D::doUpdates(GfxCanvas& can)
{
    GfxMultiClip mc(can);

    /* delete everything */
    if (need_delete) {
        container_t::iterator in = sprites.begin(), out = in;
        while (in != sprites.end()) {
            if (! (*in)->md.need_delete) {
                *out = *in;
                ++out;
            } else {
                if (! (*in)->md.is_invisible) {
                    if ((*in)->md.was_moved)
                        mc.add(expandRectangle((*in)->md.old_area));
                    else
                        mc.add(expandRectangle((*in)->pos_2d));
                }
                (*in)->setManager(0);
            }
            ++in;
        }
        sprites.erase(out, sprites.end());
        need_delete = false;
    }

    /* process sorts */
    bool need_add = true;
    if (need_sort) {
        std::stable_sort(sprites.begin(), sprites.end(), sortByDepth);
        need_sort = false;
        for (container_t::iterator i = sprites.begin(); i != sprites.end(); ++i)
            mc.add(expandRectangle((*i)->getPos()));
        need_add = false;
    }

    /* process movement and updates */
    if (need_update) {
        for (container_t::iterator i = sprites.begin(); i != sprites.end(); ++i) {
            if ((*i)->md.need_update) {
                if (need_add)
                    mc.add(expandRectangle((*i)->getPos()));
                (*i)->md.need_update = false;
            }
            if ((*i)->md.was_moved) {
                mc.add(expandRectangle((*i)->md.old_area));
                (*i)->md.was_moved = false;
            }
        }
        need_update = false;
    }

    /* redraw */
    // FIXME: GfxSpriteManager did this a bit more efficiently
    mc.clipRegionAtRectangle(bbox);

    GfxRect area = mc.getRectSet().getBoundingRectangle();
    mc.lock();
    drawBackground(mc, area);
    for (container_t::iterator i = sprites.begin(); i != sprites.end(); ++i) {
        if ((*i)->getPosZ() < 0) {
            (*i)->md.is_invisible = true;
        } else {
            (*i)->md.is_invisible = false;
            if (area.isIntersecting((*i)->getPos()))
                (*i)->draw(mc);
        }
    }
//     GfxClipRectFilter cr(area, can);
//     for (container_t::iterator i = sprites.begin(); i != sprites.end(); ++i)
//         if (area.isIntersecting((*i)->getPos()))
//             (*i)->draw(cr);
    mc.unlock();
    mc.update(area);
}

/** Handle updates without redraw. Call this multiple times if you
    wish, then redraw the whole thing using draw(). */
void
GfxSpriteManager3D::doUpdatesSilently()
{
    /* delete everything */
    if (need_delete) {
        container_t::iterator in = sprites.begin(), out = in;
        while (in != sprites.end()) {
            if (! (*in)->md.need_delete) {
                *out = *in;
                ++out;
            } else {
                (*in)->setManager(0);
            }
            ++in;
        }
        sprites.erase(out, sprites.end());
        need_delete = false;
    }

    /* process sorts */
    if (need_sort) {
        std::stable_sort(sprites.begin(), sprites.end(), sortByDepth);
        need_sort = false;
    }

    /* process movement and updates */
    for (container_t::iterator i = sprites.begin(); i != sprites.end(); ++i) {
        (*i)->md.need_update = false;
        (*i)->md.was_moved = false;
        (*i)->md.is_invisible = (*i)->getPosZ() < 0;
    }
    need_update = false;
}

/** Add sprite to this manager.
    \param sprite new sprite */
void
GfxSpriteManager3D::addSprite(Ptr<GfxSprite3D> sprite)
{
    sprites.push_back(sprite);
    sprite->setManager(this);
    sprite->md.need_update  = true;
    sprite->md.tick_mark    = tick_mark;
    sprite->md.is_invisible = sprite->getPosZ() < 0;
    need_sort = true;
    need_update = true;
}

/** Redraw everything on the given canvas. You should not call this on
    the same canvas as doUpdates() as long as there are dirty (=moved,
    resized) sprites. Calling it during clean periods or on different
    canvases is safe.
    \param can new canvas. Caller must lock and update it. */
void
GfxSpriteManager3D::draw(GfxCanvas& can)
{
    drawBackground(can, bbox);

    GfxClipRectFilter clip(bbox, can);
    for (container_t::iterator i = sprites.begin(); i != sprites.end(); ++i)
        if (!(*i)->md.need_delete && (*i)->getPosZ() >= 0)
            (*i)->draw(clip);
}

/** Draw background. Just factored out of draw() for easier handling. */
void
GfxSpriteManager3D::drawBackground(GfxCanvas& can, GfxRect area)
{
    drawSolidBar(can, area, standard_colors[COLOR_BLACK]);
}

/** Delete all sprites in a particular Z range. This will schedule all
    sprites for deletion whose Z coordinate is in the range [z0, z1).
    \param z0,z1 range. z0 is inclusive, z1 is not. */
void
GfxSpriteManager3D::deleteSpritesInZRange(int z0, int z1)
{
    for (container_t::iterator i = sprites.begin(); i != sprites.end(); ++i)
        if ((*i)->pos_z >= z0 && (*i)->pos_z < z1)
            (*i)->deleteThis();
}

/** Delete all sprites after a particular Z. This will schedule all
    sprites for deletion whose Z coordinate is z0 or higher.
    \param z0 minimum Z */
void
GfxSpriteManager3D::deleteSpritesAfterZ(int z0)
{
    for (container_t::iterator i = sprites.begin(); i != sprites.end(); ++i)
        if ((*i)->pos_z >= z0)
            (*i)->deleteThis();
}

/** Check whether we have a sprite in Z range [z0,z1).
    \param z0,z1 range. z0 is inclusive, z1 is not */
bool
GfxSpriteManager3D::haveSpritesInZRange(int z0, int z1) const
{
    for (container_t::const_iterator i = sprites.begin(); i != sprites.end(); ++i)
        if ((*i)->pos_z >= z0 && (*i)->pos_z < z1)
            return true;
    return false;
}

/** Delete all sprites with a particular Id.
    \param id Id number to delete */
void
GfxSpriteManager3D::deleteSpritesWithId(int id)
{
    for (container_t::iterator i = sprites.begin(); i != sprites.end(); ++i)
        if ((*i)->getId() == id)
            (*i)->deleteThis();
}

/** Check whether we have sprites with the specified Id.
    \param id Id number to check */
bool
GfxSpriteManager3D::haveSpritesWithId(int id) const
{
    for (container_t::const_iterator i = sprites.begin(); i != sprites.end(); ++i)
        if ((*i)->getId() == id)
            return true;
    return false;
}

/*************************** GfxPixmapSprite3D ***************************/

/*! \class GfxPixmapSprite3D

    This sprite displays a fixed pixmap; it does not move or change in
    any way. Derived classes may override tick() to add movement. */

/** Create GfxPixmapSprite with top-left position.
    \param x,y top-left corner
    \param z Z stacking order
    \param pix the pixmap */
GfxPixmapSprite3D::GfxPixmapSprite3D(int x, int y, int z, Ptr<GfxPixmap> pix)
    : GfxSprite3D(GfxRect(x, y, pix->getWidth(), pix->getHeight()), z), pix(pix)
{ }

/** Create GfxPixmapSprite with top-left position and Id.
    \param x,y top-left position
    \param z Z stacking order
    \param id identifier
    \param pix the pixmap */
GfxPixmapSprite3D::GfxPixmapSprite3D(int x, int y, int z, int id, Ptr<GfxPixmap> pix)
    : GfxSprite3D(GfxRect(x, y, pix->getWidth(), pix->getHeight()), z, id), pix(pix)
{ }

/** Create GfxPixmapSprite with center position. The bool parameter
    is ignored and exists only to make the signature unique.
    \param x,y center position
    \param z Z stacking order
    \param pix the pixmap */
GfxPixmapSprite3D::GfxPixmapSprite3D(int x, int y, int z, Ptr<GfxPixmap> pix, bool)
    : GfxSprite3D(GfxRect(x - pix->getWidth()/2, y - pix->getHeight()/2, pix->getWidth(), pix->getHeight()), z),
      pix(pix)
{ }

void
GfxPixmapSprite3D::draw(GfxCanvas& can)
{
    pix->blit(can, getPos().x, getPos().y);
}
